# Copyright (c) 2020 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

from __future__ import annotations
import typing
import struct
import dataclasses
from ipaddress import IPv4Address, IPv6Address
import pyuavcan
import pyuavcan.transport.udp
from pyuavcan.transport import Trace, TransferTrace, Capture, AlienSessionSpecifier, AlienTransferMetadata
from pyuavcan.transport import AlienTransfer, TransferFrom, Timestamp
from pyuavcan.transport.commons.high_overhead_transport import AlienTransferReassembler, TransferReassembler
from ._frame import UDPFrame
from ._ip import LinkLayerPacket, SUBJECT_PORT
from ._ip import unicast_ip_to_node_id, udp_port_to_service_data_specifier, multicast_group_to_message_data_specifier


@dataclasses.dataclass(frozen=True)
class IPPacket:
    protocol: int
    payload: memoryview

    @property
    def source_destination(
        self,
    ) -> typing.Union[typing.Tuple[IPv4Address, IPv4Address], typing.Tuple[IPv6Address, IPv6Address]]:
        raise NotImplementedError

    @staticmethod
    def parse(link_layer_packet: LinkLayerPacket) -> typing.Optional[IPPacket]:
        import socket

        if link_layer_packet.protocol == socket.AF_INET:
            return IPv4Packet.parse_payload(link_layer_packet.payload)
        if link_layer_packet.protocol == socket.AF_INET6:
            return IPv6Packet.parse_payload(link_layer_packet.payload)
        return None


@dataclasses.dataclass(frozen=True)
class IPv4Packet(IPPacket):
    source: IPv4Address
    destination: IPv4Address

    _FORMAT = struct.Struct("!BB HHH BB H II")

    def __post_init__(self) -> None:
        if self.source.is_multicast:
            raise ValueError(f"Source IP address cannot be a multicast group address")

    @property
    def source_destination(self) -> typing.Tuple[IPv4Address, IPv4Address]:
        return self.source, self.destination

    @staticmethod
    def parse_payload(link_layer_payload: memoryview) -> typing.Optional[IPv4Packet]:
        try:
            (
                ver_ihl,
                _dscp_ecn,
                total_length,
                _ident,
                _flags_frag_off,
                _ttl,
                proto,
                _hdr_chk,
                src_adr,
                dst_adr,
            ) = IPv4Packet._FORMAT.unpack_from(link_layer_payload)
        except struct.error:
            return None
        ver, ihl = ver_ihl >> 4, ver_ihl & 0xF
        if ver == 4:
            payload = link_layer_payload[ihl * 4 : total_length]
            return IPv4Packet(
                protocol=proto,
                payload=payload,
                source=IPv4Address(src_adr),
                destination=IPv4Address(dst_adr),
            )
        return None


@dataclasses.dataclass(frozen=True)
class IPv6Packet(IPPacket):
    source: IPv6Address
    destination: IPv6Address

    @property
    def source_destination(self) -> typing.Tuple[IPv6Address, IPv6Address]:
        return self.source, self.destination

    @staticmethod
    def parse_payload(link_layer_payload: memoryview) -> typing.Optional[IPv6Packet]:
        raise NotImplementedError("Support for IPv6 is not implemented yet")


@dataclasses.dataclass(frozen=True)
class UDPIPPacket:
    source_port: int
    destination_port: int
    payload: memoryview

    _FORMAT = struct.Struct("!HH HH")

    def __post_init__(self) -> None:
        if not (0 <= self.source_port <= 0xFFFF):
            raise ValueError(f"Invalid source port: {self.source_port}")
        if not (0 <= self.destination_port <= 0xFFFF):
            raise ValueError(f"Invalid destination port: {self.destination_port}")

    @staticmethod
    def parse(ip_packet: IPPacket) -> typing.Optional[UDPIPPacket]:
        if ip_packet.protocol != 0x11:  # https://en.wikipedia.org/wiki/List_of_IP_protocol_numbers
            return None
        try:
            src_port, dst_port, total_length, _udp_chk = UDPIPPacket._FORMAT.unpack_from(ip_packet.payload)
        except struct.error:
            return None
        payload = ip_packet.payload[UDPIPPacket._FORMAT.size : total_length]
        return UDPIPPacket(source_port=src_port, destination_port=dst_port, payload=payload)


@dataclasses.dataclass(frozen=True)
class UDPCapture(Capture):
    """
    The UDP transport does not differentiate between sent and received packets.
    See :meth:`pyuavcan.transport.udp.UDPTransport.begin_capture` for details.
    """

    link_layer_packet: LinkLayerPacket

    def parse(self) -> typing.Optional[typing.Tuple[pyuavcan.transport.AlienSessionSpecifier, UDPFrame]]:
        """
        The parsed representation is only defined if the packet is a valid UAVCAN/UDP frame.
        The source node-ID is never None.
        """
        ip_packet = IPPacket.parse(self.link_layer_packet)
        if ip_packet is None:
            return None
        ip_source, ip_destination = ip_packet.source_destination

        udp_packet = UDPIPPacket.parse(ip_packet)
        if udp_packet is None:
            return None

        dst_nid: typing.Optional[int]
        data_spec: typing.Optional[pyuavcan.transport.DataSpecifier]
        if ip_destination.is_multicast:
            if udp_packet.destination_port != SUBJECT_PORT:
                return None
            dst_nid = None  # Broadcast
            data_spec = multicast_group_to_message_data_specifier(ip_source, ip_destination)
        else:
            dst_nid = unicast_ip_to_node_id(ip_source, ip_destination)
            if dst_nid is None:  # The packet crosses the UAVCAN/UDP subnet boundary, invalid.
                return None
            data_spec = udp_port_to_service_data_specifier(udp_packet.destination_port)

        if data_spec is None:
            return None

        frame = UDPFrame.parse(udp_packet.payload)
        if frame is None:
            return None

        src_nid = unicast_ip_to_node_id(ip_source, ip_source)
        assert src_nid is not None
        ses_spec = pyuavcan.transport.AlienSessionSpecifier(
            source_node_id=src_nid, destination_node_id=dst_nid, data_specifier=data_spec
        )
        return ses_spec, frame

    @staticmethod
    def get_transport_type() -> typing.Type[pyuavcan.transport.udp.UDPTransport]:
        return pyuavcan.transport.udp.UDPTransport


@dataclasses.dataclass(frozen=True)
class UDPErrorTrace(pyuavcan.transport.ErrorTrace):
    error: TransferReassembler.Error


class UDPTracer(pyuavcan.transport.Tracer):
    """
    This is like a Wireshark dissector but UAVCAN-focused.
    Return types from :meth:`update`:

    - :class:`pyuavcan.transport.TransferTrace`
    - :class:`UDPErrorTrace`
    """

    def __init__(self) -> None:
        self._sessions: typing.Dict[AlienSessionSpecifier, _AlienSession] = {}

    def update(self, cap: Capture) -> typing.Optional[Trace]:
        if not isinstance(cap, UDPCapture):
            return None

        parsed = cap.parse()
        if not parsed:
            return None

        spec, frame = parsed
        return self._get_session(spec).update(cap.timestamp, frame)

    def _get_session(self, specifier: AlienSessionSpecifier) -> _AlienSession:
        try:
            return self._sessions[specifier]
        except KeyError:
            self._sessions[specifier] = _AlienSession(specifier)
        return self._sessions[specifier]


class _AlienSession:
    def __init__(self, specifier: AlienSessionSpecifier) -> None:
        assert specifier.source_node_id is not None
        self._specifier = specifier
        self._reassembler = AlienTransferReassembler(specifier.source_node_id)

    def update(self, timestamp: Timestamp, frame: UDPFrame) -> typing.Optional[Trace]:
        tid_timeout = self._reassembler.transfer_id_timeout
        tr = self._reassembler.process_frame(timestamp, frame)
        if isinstance(tr, TransferReassembler.Error):
            return UDPErrorTrace(timestamp=timestamp, error=tr)
        if isinstance(tr, TransferFrom):
            meta = AlienTransferMetadata(tr.priority, tr.transfer_id, self._specifier)
            return TransferTrace(timestamp, AlienTransfer(meta, tr.fragmented_payload), tid_timeout)
        assert tr is None
        return None


# ----------------------------------------  TESTS GO BELOW THIS LINE  ----------------------------------------


def _unittest_udp_tracer() -> None:
    import socket
    from pytest import approx
    from ipaddress import ip_address
    from pyuavcan.transport import Priority, ServiceDataSpecifier
    from pyuavcan.transport.udp import UDPTransport
    from ._ip import service_data_specifier_to_udp_port

    tr = UDPTransport.make_tracer()
    ts = Timestamp.now()
    ds = ServiceDataSpecifier(11, ServiceDataSpecifier.Role.RESPONSE)

    # VALID SERVICE FRAME
    llp = LinkLayerPacket(
        protocol=socket.AF_INET,
        source=memoryview(b""),
        destination=memoryview(b""),
        payload=memoryview(
            b"".join(
                [
                    # IPv4
                    b"\x45\x00",
                    (20 + 8 + 24 + 12).to_bytes(2, "big"),  # Total length (incl. the 20 bytes of the IP header)
                    b"\x7e\x50\x40\x00\x40",  # ID, flags, fragment offset, TTL
                    b"\x11",  # Protocol (UDP)
                    b"\x00\x00",  # IP checksum (unset)
                    ip_address("127.0.0.42").packed,  # Source
                    ip_address("127.0.0.63").packed,  # Destination
                    # UDP/IP
                    (12345).to_bytes(2, "big"),  # Source port
                    service_data_specifier_to_udp_port(ds).to_bytes(2, "big"),  # Destination port
                    (8 + 24 + 12).to_bytes(2, "big"),  # Total length (incl. the 8 bytes of the UDP header)
                    b"\x00\x00",  # UDP checksum (unset)
                    # UAVCAN/UDP
                    b"".join(
                        UDPFrame(
                            priority=Priority.SLOW,
                            transfer_id=1234567890,
                            index=0,
                            end_of_transfer=True,
                            payload=memoryview(b"Hello world!"),
                        ).compile_header_and_payload()
                    ),
                ]
            )
        ),
    )
    ip_packet = IPPacket.parse(llp)
    assert ip_packet is not None
    assert ip_packet.source_destination == (ip_address("127.0.0.42"), ip_address("127.0.0.63"))
    assert ip_packet.protocol == 0x11
    udp_packet = UDPIPPacket.parse(ip_packet)
    assert udp_packet is not None
    assert udp_packet.source_port == 12345
    assert udp_packet.destination_port == service_data_specifier_to_udp_port(ds)
    trace = tr.update(UDPCapture(ts, llp))
    assert isinstance(trace, TransferTrace)
    assert trace.timestamp == ts
    assert trace.transfer_id_timeout == approx(2.0)  # Initial value.
    assert trace.transfer.metadata.transfer_id == 1234567890
    assert trace.transfer.metadata.priority == Priority.SLOW
    assert trace.transfer.metadata.session_specifier.source_node_id == 42
    assert trace.transfer.metadata.session_specifier.destination_node_id == 63
    assert trace.transfer.metadata.session_specifier.data_specifier == ds
    assert trace.transfer.fragmented_payload == [memoryview(b"Hello world!")]

    # ANOTHER TRANSPORT, IGNORED
    assert None is tr.update(pyuavcan.transport.Capture(ts))

    # MALFORMED - UAVCAN/UDP IS EMPTY
    llp = LinkLayerPacket(
        protocol=socket.AF_INET,
        source=memoryview(b""),
        destination=memoryview(b""),
        payload=memoryview(
            b"".join(
                [
                    # IPv4
                    b"\x45\x00",
                    (20 + 8 + 24 + 12).to_bytes(2, "big"),  # Total length (incl. the 20 bytes of the IP header)
                    b"\x7e\x50\x40\x00\x40",  # ID, flags, fragment offset, TTL
                    b"\x11",  # Protocol (UDP)
                    b"\x00\x00",  # IP checksum (unset)
                    ip_address("127.0.0.42").packed,  # Source
                    ip_address("127.0.0.63").packed,  # Destination
                    # UDP/IP
                    (1).to_bytes(2, "big"),  # Source port
                    (1).to_bytes(2, "big"),  # Destination port
                    (8).to_bytes(2, "big"),  # Total length (incl. the 8 bytes of the UDP header)
                    b"\x00\x00",  # UDP checksum (unset)
                    # UAVCAN/UDP is missing
                ]
            )
        ),
    )
    ip_packet = IPPacket.parse(llp)
    assert ip_packet is not None
    assert ip_packet.source_destination == (ip_address("127.0.0.42"), ip_address("127.0.0.63"))
    assert ip_packet.protocol == 0x11
    udp_packet = UDPIPPacket.parse(ip_packet)
    assert udp_packet is not None
    assert udp_packet.source_port == 1
    assert udp_packet.destination_port == 1
    assert None is tr.update(UDPCapture(ts, llp))



# Copyright (c) 2019 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

from __future__ import annotations
import typing
import dataclasses
import pyuavcan.util
from .media import DataFrame, FrameFormat


TRANSFER_ID_MODULO = 32

TRANSFER_CRC_LENGTH_BYTES = 2


@dataclasses.dataclass(frozen=True)
class UAVCANFrame:
    identifier: int
    transfer_id: int
    start_of_transfer: bool
    end_of_transfer: bool
    toggle_bit: bool
    padded_payload: memoryview

    def __post_init__(self) -> None:
        if self.transfer_id < 0:
            raise ValueError("Transfer ID cannot be negative")

        if self.start_of_transfer and not self.toggle_bit:
            raise ValueError(f"The toggle bit must be set in the first frame of the transfer")

    def compile(self) -> DataFrame:
        tail = self.transfer_id % TRANSFER_ID_MODULO
        if self.start_of_transfer:
            tail |= 1 << 7
        if self.end_of_transfer:
            tail |= 1 << 6
        if self.toggle_bit:
            tail |= 1 << 5

        data = bytearray(self.padded_payload)
        data.append(tail)
        return DataFrame(FrameFormat.EXTENDED, self.identifier, data)

    @staticmethod
    def parse(source: DataFrame) -> typing.Optional[UAVCANFrame]:
        if source.format != FrameFormat.EXTENDED:
            return None
        if len(source.data) < 1:
            return None

        padded_payload, tail = memoryview(source.data)[:-1], source.data[-1]
        transfer_id = tail & (TRANSFER_ID_MODULO - 1)
        sot, eot, tog = tuple(tail & (1 << x) != 0 for x in (7, 6, 5))
        if sot and not tog:
            return None

        return UAVCANFrame(
            identifier=source.identifier,
            transfer_id=transfer_id,
            start_of_transfer=sot,
            end_of_transfer=eot,
            toggle_bit=tog,
            padded_payload=padded_payload,
        )

    @staticmethod
    def get_required_padding(data_length: int) -> int:
        return DataFrame.get_required_padding(data_length + 1)  # +1 for the tail byte

    def __repr__(self) -> str:
        kwargs = {f.name: getattr(self, f.name) for f in dataclasses.fields(self)}
        kwargs["identifier"] = f"0x{self.identifier:08x}"
        kwargs["padded_payload"] = bytes(self.padded_payload).hex()
        return pyuavcan.util.repr_attributes(self, **kwargs)


def compute_transfer_id_forward_distance(a: int, b: int) -> int:
    """
    The algorithm is defined in the CAN bus transport layer specification of the UAVCAN Specification.
    """
    assert a >= 0 and b >= 0
    a %= TRANSFER_ID_MODULO
    b %= TRANSFER_ID_MODULO
    d = b - a
    if d < 0:
        d += TRANSFER_ID_MODULO

    assert 0 <= d < TRANSFER_ID_MODULO
    assert (a + d) & (TRANSFER_ID_MODULO - 1) == b
    return d


def _unittest_can_transfer_id_forward_distance() -> None:
    cfd = compute_transfer_id_forward_distance
    assert 0 == cfd(0, 0)
    assert 1 == cfd(0, 1)
    assert 7 == cfd(0, 7)
    assert 0 == cfd(7, 7)
    assert 1 == cfd(31, 0)
    assert 5 == cfd(0, 5)
    assert 31 == cfd(31, 30)
    assert 30 == cfd(7, 5)


def _unittest_can_uavcan_frame() -> None:
    from pytest import raises

    UAVCANFrame(123, 123, True, False, True, memoryview(b""))
    UAVCANFrame(123, 123, False, False, True, memoryview(b""))
    UAVCANFrame(123, 123, False, False, False, memoryview(b""))

    with raises(ValueError):
        UAVCANFrame(123, -1, True, False, True, memoryview(b""))

    with raises(ValueError):
        UAVCANFrame(123, 123, True, False, False, memoryview(b""))

    ref = UAVCANFrame(
        identifier=0,
        transfer_id=0,
        start_of_transfer=False,
        end_of_transfer=False,
        toggle_bit=False,
        padded_payload=memoryview(b""),
    )
    assert ref == UAVCANFrame.parse(DataFrame(FrameFormat.EXTENDED, 0, bytearray(b"\x00")))

    ref = UAVCANFrame(
        identifier=123456,
        transfer_id=12,
        start_of_transfer=True,
        end_of_transfer=False,
        toggle_bit=True,
        padded_payload=memoryview(b"Hello"),
    )
    assert ref == UAVCANFrame.parse(DataFrame(FrameFormat.EXTENDED, 123456, bytearray(b"Hello\xAC")))

    ref = UAVCANFrame(
        identifier=1234567,
        transfer_id=12,
        start_of_transfer=False,
        end_of_transfer=True,
        toggle_bit=True,
        padded_payload=memoryview(b"Hello"),
    )
    assert ref == UAVCANFrame.parse(DataFrame(FrameFormat.EXTENDED, 1234567, bytearray(b"Hello\x6C")))

    assert UAVCANFrame.parse(DataFrame(FrameFormat.EXTENDED, 1234567, bytearray(b"Hello\xCC"))) is None  # Bad toggle

    assert UAVCANFrame.parse(DataFrame(FrameFormat.EXTENDED, 1234567, bytearray(b""))) is None  # No tail byte

    assert UAVCANFrame.parse(DataFrame(FrameFormat.BASE, 123, bytearray(b"Hello\x6C"))) is None  # Bad frame format



# Copyright (c) 2019 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

import typing
import asyncio
import logging
import dataclasses
import pyuavcan.transport
from ._session import RedundantInputSession, RedundantOutputSession, RedundantSession
from ._error import InconsistentInferiorConfigurationError
from ._deduplicator import Deduplicator
from ._tracer import RedundantTracer, RedundantCapture


_logger = logging.getLogger(__name__)


@dataclasses.dataclass
class RedundantTransportStatistics(pyuavcan.transport.TransportStatistics):
    """
    Aggregate statistics for all inferior transports in a redundant group.
    This is an atomic immutable sample; it is not updated after construction.
    """

    inferiors: typing.List[pyuavcan.transport.TransportStatistics] = dataclasses.field(default_factory=list)
    """
    The ordering is guaranteed to match that of :attr:`RedundantTransport.inferiors`.
    """


class RedundantTransport(pyuavcan.transport.Transport):
    """
    This is a composite over a set of :class:`pyuavcan.transport.Transport`.
    Please read the module documentation for details.
    """

    def __init__(self, *, loop: typing.Optional[asyncio.AbstractEventLoop] = None) -> None:
        """
        :param loop: All inferiors shall run on the same event loop,
            which is configured once here and cannot be changed after the instance is constructed.
            If not provided, defaults to :func:`asyncio.get_event_loop`.
        """
        self._cols: typing.List[pyuavcan.transport.Transport] = []
        self._rows: typing.Dict[pyuavcan.transport.SessionSpecifier, RedundantSession] = {}
        self._unwrapped_capture_handlers: typing.List[typing.Callable[[RedundantCapture], None]] = []
        self._loop = loop if loop is not None else asyncio.get_event_loop()
        self._check_matrix_consistency()

    @property
    def protocol_parameters(self) -> pyuavcan.transport.ProtocolParameters:
        """
        Aggregate parameters constructed from all inferiors.
        If there are no inferiors (i.e., if the instance is closed), the value is all-zeros.
        Beware that if the set of inferiors is changed, this value may also be changed.

        The values are obtained from the set of inferiors by applying the following reductions:

        - min transfer-ID modulo
        - min max-nodes
        - min MTU
        """
        ipp = [t.protocol_parameters for t in self._cols] or [
            pyuavcan.transport.ProtocolParameters(
                transfer_id_modulo=0,
                max_nodes=0,
                mtu=0,
            )
        ]
        return pyuavcan.transport.ProtocolParameters(
            transfer_id_modulo=min(t.transfer_id_modulo for t in ipp),
            max_nodes=min(t.max_nodes for t in ipp),
            mtu=min(t.mtu for t in ipp),
        )

    @property
    def loop(self) -> asyncio.AbstractEventLoop:
        """
        All inferiors run on the same event loop, which is configured statically once when the redundant transport
        is instantiated.
        The loop cannot be reassigned after instantiation.
        """
        return self._loop

    @property
    def local_node_id(self) -> typing.Optional[int]:
        """
        All inferiors share the same local node-ID.
        If there are no inferiors, the value is None (anonymous).
        """
        if self._cols:
            nid_set = set(x.local_node_id for x in self._cols)
            if len(nid_set) == 1:
                (out,) = nid_set
                return out
            # The following exception should not occur during normal operation unless one of the inferiors is
            # reconfigured sneakily.
            raise InconsistentInferiorConfigurationError(
                f"Redundant transports have different node-IDs: {[x.local_node_id for x in self._cols]}"
            )
        return None

    def get_input_session(
        self, specifier: pyuavcan.transport.InputSessionSpecifier, payload_metadata: pyuavcan.transport.PayloadMetadata
    ) -> RedundantInputSession:
        out = self._get_session(
            specifier,
            lambda fin: RedundantInputSession(
                specifier, payload_metadata, lambda: self.protocol_parameters.transfer_id_modulo, self._loop, fin
            ),
        )
        assert isinstance(out, RedundantInputSession)
        self._check_matrix_consistency()
        return out

    def get_output_session(
        self, specifier: pyuavcan.transport.OutputSessionSpecifier, payload_metadata: pyuavcan.transport.PayloadMetadata
    ) -> RedundantOutputSession:
        out = self._get_session(
            specifier, lambda fin: RedundantOutputSession(specifier, payload_metadata, self._loop, fin)
        )
        assert isinstance(out, RedundantOutputSession)
        self._check_matrix_consistency()
        return out

    def sample_statistics(self) -> RedundantTransportStatistics:
        return RedundantTransportStatistics(inferiors=[t.sample_statistics() for t in self._cols])

    @property
    def input_sessions(self) -> typing.Sequence[RedundantInputSession]:
        return [s for s in self._rows.values() if isinstance(s, RedundantInputSession)]

    @property
    def output_sessions(self) -> typing.Sequence[RedundantOutputSession]:
        return [s for s in self._rows.values() if isinstance(s, RedundantOutputSession)]

    @property
    def inferiors(self) -> typing.Sequence[pyuavcan.transport.Transport]:
        """
        Read-only access to the list of inferior transports.
        The inferiors are guaranteed to be ordered according to the temporal order of their attachment.
        """
        return self._cols[:]  # Return copy to prevent mutation

    def attach_inferior(self, transport: pyuavcan.transport.Transport) -> None:
        """
        Adds a new transport to the redundant group. The new transport shall not be closed.

        If the transport is already added or it is the redundant transport itself (recursive attachment),
        a :class:`ValueError` will be raised.

        If the configuration of the new transport is not compatible with the other inferiors or with the
        redundant transport instance itself, an instance of :class:`InconsistentInferiorConfigurationError`
        will be raised.
        Specifically, the following preconditions are checked:

        - The new inferior shall operate on the same event loop as the redundant transport instance it is added to.
        - The local node-ID shall be the same for all inferiors, or all shall be anonymous.
        - The transfer-ID modulo shall meet *either* of the following conditions:

            - Identical for all inferiors.
            - Not less than :attr:`MONOTONIC_TRANSFER_ID_MODULO_THRESHOLD` for all inferiors.

        If an exception is raised while the setup of the new inferior is in progress,
        the operation will be rolled back to ensure state consistency.
        """
        self._validate_inferior(transport)
        self._cols.append(transport)
        try:
            for redundant_session in self._rows.values():
                self._construct_inferior_session(transport, redundant_session)
        except Exception:
            self.detach_inferior(transport)  # Roll back to ensure consistent states.
            raise
        finally:
            self._check_matrix_consistency()
        # Launch the capture as late as possible to not leave it dangling if the attachment failed.
        for ch in self._unwrapped_capture_handlers:
            transport.begin_capture(self._wrap_capture_handler(transport, ch))

    def detach_inferior(self, transport: pyuavcan.transport.Transport) -> None:
        """
        Removes the specified transport from the redundant group.
        If there is no such transport, a :class:`ValueError` will be raised.

        All sessions of the removed inferior that are managed by the redundant transport instance
        will be automatically closed, but the inferior itself will not be
        (the caller will have to do that manually if desired).
        """
        if transport not in self._cols:
            raise ValueError(f"{transport} is not an inferior of {self}")
        index = self._cols.index(transport)
        self._cols.remove(transport)
        for owner in self._rows.values():
            try:
                owner._close_inferior(index)  # pylint: disable=protected-access
            except Exception as ex:
                _logger.exception("%s could not close inferior session #%d in %s: %s", self, index, owner, ex)
        self._check_matrix_consistency()

    def close(self) -> None:
        """
        Closes all redundant session instances, detaches and closes all inferior transports.
        Any exceptions occurring in the process will be suppressed and logged.

        Upon completion, the session matrix will be returned into its original empty state.
        It can be populated back by adding new transports and/or instantiating new redundant sessions
        if needed.
        In other words, closing is reversible here, which is uncommon for the library;
        consider this feature experimental.

        If the session matrix is empty, this method has no effect.
        """
        for s in list(self._rows.values()):
            try:
                s.close()
            except Exception as ex:  # pragma: no cover
                _logger.exception("%s could not close %s: %s", self, s, ex)

        for t in self._cols:
            try:
                t.close()
            except Exception as ex:  # pragma: no cover
                _logger.exception("%s could not close inferior %s: %s", self, t, ex)

        self._cols.clear()
        assert not self._rows, "All sessions should have been unregistered"
        self._check_matrix_consistency()

    def begin_capture(self, handler: pyuavcan.transport.CaptureCallback) -> None:
        """
        Stores the handler in the local list of handlers.
        Invokes :class:`pyuavcan.transport.Transport.begin_capture` on each inferior.
        If at least one inferior raises an exception, it is propagated immediately and the remaining inferiors
        will remain in an inconsistent state.
        When a new inferior is added later, the stored handlers will be automatically used to enable capture on it.
        If such auto-restoration behavior is undesirable, configure capture individually per-inferior instead.

        Every capture emitted by the inferiors is wrapped into :class:`RedundantCapture`,
        which contains additional metadata about the inferior transport instance that emitted the capture.
        This is done to let users understand which transport of the redundant group has
        provided the capture and also this information is used by :class:`RedundantTracer`
        to automatically manage transfer deduplication.
        """
        self._unwrapped_capture_handlers.append(handler)
        for c in self._cols:
            c.begin_capture(self._wrap_capture_handler(c, handler))

    @property
    def capture_active(self) -> bool:
        return len(self._unwrapped_capture_handlers) > 0

    @staticmethod
    def make_tracer() -> RedundantTracer:
        """
        See :class:`RedundantTracer`.
        """
        return RedundantTracer()

    async def spoof(self, transfer: pyuavcan.transport.AlienTransfer, monotonic_deadline: float) -> bool:
        """
        Simply propagates the call to every inferior.
        The return value is a logical AND for all inferiors; False if there are no inferiors.

        First exception to occur terminates the operation and is raised immediately.
        This is different from regular sending; the assumption is that the caller necessarily wants to ensure
        that spoofing takes place against every inferior.
        If this is not the case, spoof each inferior separately.
        """
        if not self._cols:
            return False
        gather = asyncio.gather(*[inf.spoof(transfer, monotonic_deadline) for inf in self._cols])
        try:
            results = await gather
        except Exception:
            gather.cancel()
            raise
        return all(results)

    def _validate_inferior(self, transport: pyuavcan.transport.Transport) -> None:
        # Ensure all inferiors run on the same event loop.
        if self.loop is not transport.loop:
            raise InconsistentInferiorConfigurationError(
                f"The inferior operates on a different event loop {transport.loop}, expected {self.loop}"
            )

        # Prevent double-add.
        if transport in self._cols:
            raise ValueError(f"{transport} is already an inferior of {self}")

        # Just out of abundance of paranoia.
        if transport is self:
            raise ValueError(f"A redundant transport cannot be an inferior of itself")

        # If there are no other inferiors, no further checks are necessary.
        if self._cols:
            # Ensure all inferiors have the same node-ID.
            if self.local_node_id != transport.local_node_id:
                raise InconsistentInferiorConfigurationError(
                    f"The inferior has a different node-ID {transport.local_node_id}, expected {self.local_node_id}"
                )

            # Ensure all inferiors use the same transfer-ID overflow policy.
            if self.protocol_parameters.transfer_id_modulo >= Deduplicator.MONOTONIC_TRANSFER_ID_MODULO_THRESHOLD:
                if (
                    transport.protocol_parameters.transfer_id_modulo
                    < Deduplicator.MONOTONIC_TRANSFER_ID_MODULO_THRESHOLD
                ):
                    raise InconsistentInferiorConfigurationError(
                        f"The new inferior shall use monotonic transfer-ID counters in order to match the "
                        f"other inferiors in the redundant transport group"
                    )
            else:
                tid_modulo = self.protocol_parameters.transfer_id_modulo
                if transport.protocol_parameters.transfer_id_modulo != tid_modulo:
                    raise InconsistentInferiorConfigurationError(
                        f"The transfer-ID modulo {transport.protocol_parameters.transfer_id_modulo} of the new "
                        f"inferior is not compatible with the other inferiors ({tid_modulo})"
                    )

    def _get_session(
        self,
        specifier: pyuavcan.transport.SessionSpecifier,
        session_factory: typing.Callable[[typing.Callable[[], None]], RedundantSession],
    ) -> RedundantSession:
        if specifier not in self._rows:

            def retire() -> None:
                try:
                    del self._rows[specifier]
                except LookupError:
                    pass

            ses = session_factory(retire)
            try:
                for t in self._cols:
                    self._construct_inferior_session(t, ses)
            except Exception:
                ses.close()
                raise
            assert specifier not in self._rows
            self._rows[specifier] = ses

        return self._rows[specifier]

    @staticmethod
    def _construct_inferior_session(transport: pyuavcan.transport.Transport, owner: RedundantSession) -> None:
        assert isinstance(transport, pyuavcan.transport.Transport)
        if isinstance(owner, pyuavcan.transport.InputSession):
            inferior = transport.get_input_session(owner.specifier, owner.payload_metadata)
        elif isinstance(owner, pyuavcan.transport.OutputSession):
            inferior = transport.get_output_session(owner.specifier, owner.payload_metadata)
        else:
            assert False
        assert isinstance(owner, RedundantSession)  # MyPy makes me miss static typing so much.
        # If anything whatsoever goes wrong, just roll everything back and re-raise the exception.
        new_index = len(owner.inferiors)
        try:
            owner._add_inferior(inferior)  # pylint: disable=protected-access
        except Exception:
            # The inferior MUST be closed manually because in the case of failure it is not registered
            # in the redundant session.
            inferior.close()
            # If the inferior has not been added, this method will have no effect:
            owner._close_inferior(new_index)  # pylint: disable=protected-access
            raise

    def _check_matrix_consistency(self) -> None:
        for row in self._rows.values():
            assert len(row.inferiors) == len(self._cols)

    def _wrap_capture_handler(
        self,
        inferior: pyuavcan.transport.Transport,
        handler: typing.Callable[[RedundantCapture], None],
    ) -> pyuavcan.transport.CaptureCallback:
        # If you are reading this, send me a postcard.
        return lambda cap: handler(
            RedundantCapture(
                cap.timestamp,
                inferior=cap,
                iface_id=id(inferior),
                transfer_id_modulo=self.protocol_parameters.transfer_id_modulo,  # THIS IS PROBABLY SLOW?
            )
        )

    def _get_repr_fields(self) -> typing.Tuple[typing.List[typing.Any], typing.Dict[str, typing.Any]]:
        return list(self.inferiors), {}



# Copyright (c) 2021 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

from __future__ import annotations
import typing
import logging
import dataclasses
import pyuavcan
import pyuavcan.transport.redundant
from ._deduplicator import Deduplicator


@dataclasses.dataclass(frozen=True)
class RedundantCapture(pyuavcan.transport.Capture):
    """
    Composes :class:`pyuavcan.transport.Capture` with a reference to the
    transport instance that yielded this capture.
    The user may construct such captures manually when performing postmortem analysis of a network data dump
    to feed them later into :class:`RedundantTracer`.
    """

    inferior: pyuavcan.transport.Capture
    """
    The original capture from the inferior transport.
    """

    iface_id: int
    """
    A unique number that identifies this transport in its redundant group.
    """

    transfer_id_modulo: int
    """
    The number of unique transfer-ID values (that is, the maximum possible transfer-ID plus one)
    for the transport that emitted this capture.
    This is actually a transport-specific constant.
    This value is used by :class:`RedundantTracer` to select the appropriate transfer deduplication strategy.
    """

    @staticmethod
    def get_transport_type() -> typing.Type[pyuavcan.transport.redundant.RedundantTransport]:
        return pyuavcan.transport.redundant.RedundantTransport


@dataclasses.dataclass(frozen=True)
class RedundantDuplicateTransferTrace(pyuavcan.transport.Trace):
    """
    Indicates that the last capture object completed a valid transfer that was discarded as a duplicate
    (either received from another redundant interface or deterministic data loss mitigation (DDLM) is employed).

    Observe that it is NOT a subclass of :class:`pyuavcan.transport.TransferTrace`!
    It shall not be one because duplicates should not be processed normally.
    """


class RedundantTracer(pyuavcan.transport.Tracer):
    """
    The redundant tracer automatically deduplicates transfers received from multiple redundant transports.
    It can be used either in real-time or during postmortem analysis.
    In the latter case the user would construct instances of :class:`RedundantCapture` manually and feed them
    into the tracer one-by-one.
    """

    def __init__(self) -> None:
        self._deduplicators: typing.Dict[RedundantTracer._DeduplicatorSelector, Deduplicator] = {}
        self._last_transfer_id_modulo = 0
        self._inferior_tracers: typing.Dict[
            typing.Tuple[typing.Type[pyuavcan.transport.Transport], int],
            pyuavcan.transport.Tracer,
        ] = {}

    def update(self, cap: pyuavcan.transport.Capture) -> typing.Optional[pyuavcan.transport.Trace]:
        """
        All instances of :class:`pyuavcan.transport.TransferTrace` are deduplicated,
        duplicates are simply dropped and :class:`RedundantDuplicateTransferTrace` is returned.
        All other instances (such as :class:`pyuavcan.transport.ErrorTrace`) are returned unchanged.
        """
        _logger.debug("%r: Processing %r", self, cap)
        if not isinstance(cap, RedundantCapture):
            return None

        if cap.transfer_id_modulo != self._last_transfer_id_modulo:
            _logger.info(
                "%r: TID modulo change detected, resetting state (%d deduplicators dropped): %r --> %r",
                self,
                len(self._deduplicators),
                self._last_transfer_id_modulo,
                cap.transfer_id_modulo,
            )
            # Should we also drop the tracers here? If an inferior transport is removed its tracer will be sitting
            # here useless, we don't want that. But on the other hand, disturbing the state too much is also no good.
            self._last_transfer_id_modulo = cap.transfer_id_modulo
            self._deduplicators.clear()

        tracer = self._get_inferior_tracer(cap.inferior.get_transport_type(), cap.iface_id)
        trace = tracer.update(cap.inferior)
        if not isinstance(trace, pyuavcan.transport.TransferTrace):
            _logger.debug("%r: BYPASS: %r", self, trace)
            return trace

        meta = trace.transfer.metadata
        deduplicator = self._get_deduplicator(
            meta.session_specifier.destination_node_id,
            meta.session_specifier.data_specifier,
            cap.transfer_id_modulo,
        )
        should_accept = deduplicator.should_accept_transfer(
            iface_id=cap.iface_id,
            transfer_id_timeout=trace.transfer_id_timeout,
            timestamp=trace.timestamp,
            source_node_id=meta.session_specifier.source_node_id,
            transfer_id=meta.transfer_id,
        )
        if should_accept:
            _logger.debug("%r: ACCEPT: %r", self, trace)
            return trace
        _logger.debug("%r: REJECT: %r", self, trace)
        return RedundantDuplicateTransferTrace(cap.timestamp)

    def _get_deduplicator(
        self,
        destination_node_id: typing.Optional[int],
        data_specifier: pyuavcan.transport.DataSpecifier,
        transfer_id_modulo: int,
    ) -> Deduplicator:
        selector = RedundantTracer._DeduplicatorSelector(destination_node_id, data_specifier)
        try:
            return self._deduplicators[selector]
        except LookupError:
            dd = Deduplicator.new(transfer_id_modulo)
            _logger.debug("%r: New deduplicator for %r: %r", self, selector, dd)
            self._deduplicators[selector] = dd
        return self._deduplicators[selector]

    def _get_inferior_tracer(
        self,
        inferior_type: typing.Type[pyuavcan.transport.Transport],
        inferior_iface_id: int,
    ) -> pyuavcan.transport.Tracer:
        selector = inferior_type, inferior_iface_id
        try:
            return self._inferior_tracers[selector]
        except LookupError:
            it = inferior_type.make_tracer()
            _logger.debug("%r: New inferior tracer for %r: %r", self, selector, it)
            self._inferior_tracers[selector] = it
        return self._inferior_tracers[selector]

    @dataclasses.dataclass(frozen=True)
    class _DeduplicatorSelector:
        destination_node_id: typing.Optional[int]
        data_specifier: pyuavcan.transport.DataSpecifier

    def __repr__(self) -> str:
        return pyuavcan.util.repr_attributes(self, self._inferior_tracers)


_logger = logging.getLogger(__name__)



# Copyright (c) 2019 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

import pyuavcan.transport


class InconsistentInferiorConfigurationError(pyuavcan.transport.InvalidTransportConfigurationError):
    """
    Raised when a redundant transport instance is asked to attach a new inferior whose configuration
    does not match that of the other inferiors or of the redundant transport itself.
    """



# Copyright (c) 2019 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

from __future__ import annotations
import abc
import typing
import asyncio
import dataclasses
import pyuavcan.util
from ._session import InputSession, OutputSession, InputSessionSpecifier, OutputSessionSpecifier
from ._payload_metadata import PayloadMetadata
from ._tracer import CaptureCallback, Tracer, AlienTransfer


@dataclasses.dataclass(frozen=True)
class ProtocolParameters:
    """
    Basic transport capabilities. These parameters are defined by the underlying transport specifications.

    Normally, the values should never change for a particular transport instance.
    This is not a hard guarantee, however.
    For example, a redundant transport aggregator may return a different set of parameters after
    the set of aggregated transports is changed (i.e., a transport is added or removed).
    """

    transfer_id_modulo: int
    """
    The cardinality of the set of distinct transfer-ID values; i.e., the overflow period.
    All high-overhead transports (UDP, Serial, etc.) use a sufficiently large value that will never overflow
    in a realistic, practical scenario.
    The background and motivation are explained at https://forum.uavcan.org/t/alternative-transport-protocols/324.
    Example: 32 for CAN, (2**64) for UDP.
    """

    max_nodes: int
    """
    How many nodes can the transport accommodate in a given network.
    Example: 128 for CAN, 65535 for UDP.
    """

    mtu: int
    """
    The maximum number of payload bytes in a single-frame transfer.
    If the number of payload bytes in a transfer exceeds this limit, the transport will spill
    the data into a multi-frame transfer.
    Example: 7 for Classic CAN, <=63 for CAN FD.
    """


@dataclasses.dataclass
class TransportStatistics:
    """
    Base class for transport-specific low-level statistical counters.
    Not to be confused with :class:`pyuavcan.transport.SessionStatistics`,
    which is tracked per-session.
    """


class Transport(abc.ABC):
    """
    An abstract UAVCAN transport interface. Please read the module documentation for details.

    Implementations should ensure that properties do not raise exceptions.
    """

    @property
    @abc.abstractmethod
    def loop(self) -> asyncio.AbstractEventLoop:
        """
        The asyncio event loop used to operate the transport instance.
        """
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def protocol_parameters(self) -> ProtocolParameters:
        """
        Provides information about the properties of the transport protocol implemented by the instance.
        See :class:`ProtocolParameters`.
        """
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def local_node_id(self) -> typing.Optional[int]:
        """
        The node-ID is set once during initialization of the transport,
        either explicitly (e.g., CAN) or by deriving the node-ID value from the configuration
        of the underlying protocol layers (e.g., UDP/IP).

        If the transport does not have a node-ID, this property has the value of None,
        and the transport (and the node that uses it) is said to be in the anonymous mode.
        While in the anonymous mode, some transports may choose to operate in a particular regime to facilitate
        plug-and-play node-ID allocation (for example, a CAN transport may disable automatic retransmission).

        Protip: If you feel like assigning the node-ID after initialization,
        make a proxy that implements this interface and keeps a private transport instance.
        When the node-ID is assigned, the private transport instance is destroyed,
        a new one is implicitly created in its place, and all of the dependent session instances are automatically
        recreated transparently for the user of the proxy.
        This logic is implemented in the redundant transport, which can be used even if no redundancy is needed.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def close(self) -> None:
        """
        Closes all active sessions, underlying media instances, and other resources related to this transport instance.

        After a transport is closed, none of its methods nor dependent objects (such as sessions) can be used.
        Methods invoked on a closed transport or any of its dependent objects should immediately
        raise :class:`pyuavcan.transport.ResourceClosedError`.
        Subsequent calls to close() will have no effect.

        Failure to close any of the resources does not prevent the method from closing other resources
        (best effort policy).
        Related exceptions may be suppressed and logged; the last occurred exception may be raised after
        all resources are closed if such behavior is considered to be meaningful.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def get_input_session(self, specifier: InputSessionSpecifier, payload_metadata: PayloadMetadata) -> InputSession:
        """
        This factory method is the only valid way of constructing input session instances.
        Beware that construction and retirement of sessions may be costly.

        The transport will always return the same instance unless there is no session object with the requested
        specifier, in which case it will be created and stored internally until closed.
        The payload metadata parameter is used only when a new instance is created, ignored otherwise.
        Implementations are encouraged to use a covariant return type annotation.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def get_output_session(self, specifier: OutputSessionSpecifier, payload_metadata: PayloadMetadata) -> OutputSession:
        """
        This factory method is the only valid way of constructing output session instances.
        Beware that construction and retirement of sessions may be costly.

        The transport will always return the same instance unless there is no session object with the requested
        specifier, in which case it will be created and stored internally until closed.
        The payload metadata parameter is used only when a new instance is created, ignored otherwise.
        Implementations are encouraged to use a covariant return type annotation.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def sample_statistics(self) -> TransportStatistics:
        """
        Samples the low-level transport stats.
        The returned object shall be new or cloned (should not refer to an internal field).
        Implementations should annotate the return type as a derived custom type.
        """
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def input_sessions(self) -> typing.Sequence[InputSession]:
        """
        Immutable view of all input sessions that are currently open.
        """
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def output_sessions(self) -> typing.Sequence[OutputSession]:
        """
        Immutable view of all output sessions that are currently open.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def begin_capture(self, handler: CaptureCallback) -> None:
        """
        .. warning::
            This API entity is not yet stable. Suggestions and feedback are welcomed at https://forum.uavcan.org.

        Activates low-level monitoring of the transport interface.
        Also see related method :meth:`make_tracer`.

        This method puts the transport instance into the low-level capture mode which does not interfere with its
        normal operation but may significantly increase the computing load due to the need to process every frame
        exchanged over the network (not just frames that originate or terminate at the local node).
        This usually involves reconfiguration of the local networking hardware.
        For instance, the network card may be put into promiscuous mode,
        the CAN adapter will have its acceptance filters disabled, etc.

        The capture handler is invoked for every transmitted or received transport frame and, possibly, some
        additional transport-implementation-specific events (e.g., network errors or hardware state changes)
        which are described in the specific transport implementation docs.
        The temporal order of the events delivered to the user may be distorted, depending on the guarantees
        provided by the hardware and its driver.
        This means that if the network hardware sees TX frame A and then RX frame B separated by a very short time
        interval, the user may occasionally see the sequence inverted as (B, A).

        There may be an arbitrary number of capture handlers installed; when a new handler is installed, it is
        added to the existing ones, if any.

        If the transport does not support capture, this method may have no observable effect.
        Technically, the capture protocol, as you can see, does not present any requirements to the emitted events,
        so an implementation that pretends to enter the capture mode while not actually doing anything is compliant.

        Since capture reflects actual network events, deterministic data loss mitigation will make the instance emit
        duplicate frames for affected transfers (although this is probably obvious enough without this elaboration).

        It is not possible to disable capture. Once enabled, it will go on until the transport instance is destroyed.

        :param handler: A one-argument callable invoked to inform the user about low-level network events.
            The type of the argument is :class:`Capture`, see transport-specific docs for the list of the possible
            concrete types and what events they represent.
            **The handler may be invoked from a different thread so the user should ensure synchronization.**
            If the handler raises an exception, it is suppressed and logged.
        """
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def capture_active(self) -> bool:
        """
        Whether :meth:`begin_capture` was invoked and packet capture is being performed on this transport.
        """
        raise NotImplementedError

    @staticmethod
    @abc.abstractmethod
    def make_tracer() -> Tracer:
        """
        .. warning::
            This API entity is not yet stable. Suggestions and feedback are welcomed at https://forum.uavcan.org.

        Use this factory method for constructing tracer implementations for specific transports.
        Concrete tracers may be Voldemort types themselves.
        See also: :class:`Tracer`, :meth:`begin_capture`.
        """
        raise NotImplementedError

    @abc.abstractmethod
    async def spoof(self, transfer: AlienTransfer, monotonic_deadline: float) -> bool:
        """
        .. warning::
            This API entity is not yet stable. Suggestions and feedback are welcomed at https://forum.uavcan.org.

        Send a spoofed transfer to the network.
        The configuration of the local transport instance has no effect on spoofed transfers;
        as such, even anonymous instances may send arbitrary spoofed transfers.
        The only relevant property of the instance is which network interface to use for spoofing.

        When this method is invoked for the first time, the transport instance may need to perform one-time
        initialization such as reconfiguring the networking hardware or loading additional drivers.
        Once this one-time initialization is performed,
        the transport instance will reside in the spoofing mode until the instance is closed;
        it is not possible to leave the spoofing mode without closing the instance.
        Some transports/platforms may require special permissions to perform spoofing (esp. IP-based transports).

        If the source node-ID is not provided, an anonymous transfer will be emitted.
        If anonymous transfers are not supported, :class:`pyuavcan.transport.OperationNotDefinedForAnonymousNodeError`
        will be raised.
        Same will happen if one attempted to transmit a multi-frame anonymous transfer.

        If the destination node-ID is not provided, a broadcast transfer will be emitted.
        If the data specifier is that of a service, a :class:`UnsupportedSessionConfigurationError` will be raised.
        The reverse conflict for messages is handled identically.

        Transports with cyclic transfer-ID will compute the modulo automatically.

        This method will update the appropriate statistical counters as usual.

        :returns: True on success, False on timeout.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def _get_repr_fields(self) -> typing.Tuple[typing.List[typing.Any], typing.Dict[str, typing.Any]]:
        """
        Returns a list of positional and keyword arguments to :func:`pyuavcan.util.repr_attributes_noexcept`
        for processing the :meth:`__repr__` call.
        The resulting string constructed by repr should resemble a valid Python expression that would yield
        an identical transport instance upon its evaluation.
        """
        raise NotImplementedError

    def __repr__(self) -> str:
        """
        Implementations should never override this method. Instead, see :meth:`_get_repr_fields`.
        """
        positional, keyword = self._get_repr_fields()
        return pyuavcan.util.repr_attributes_noexcept(self, *positional, **keyword)



# Copyright (c) 2019 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

"""
Publishes ``uavcan.node.Heartbeat`` periodically and provides a couple of basic auxiliary services;
see :class:`pyuavcan.application.heartbeat_publisher.HeartbeatPublisher`.
"""

from __future__ import annotations
import enum
import time
import typing
import logging
import asyncio
import uavcan.node
from uavcan.node import Heartbeat_1_0 as Heartbeat
import pyuavcan
import pyuavcan.application


class Health(enum.IntEnum):
    """
    Mirrors the health enumeration defined in ``uavcan.node.Heartbeat``.
    When enumerations are natively supported in DSDL, this will be replaced with an alias.
    """

    NOMINAL = uavcan.node.Health_1_0.NOMINAL
    ADVISORY = uavcan.node.Health_1_0.ADVISORY
    CAUTION = uavcan.node.Health_1_0.CAUTION
    WARNING = uavcan.node.Health_1_0.WARNING


class Mode(enum.IntEnum):
    """
    Mirrors the mode enumeration defined in ``uavcan.node.Heartbeat``.
    When enumerations are natively supported in DSDL, this will be replaced with an alias.
    """

    OPERATIONAL = uavcan.node.Mode_1_0.OPERATIONAL
    INITIALIZATION = uavcan.node.Mode_1_0.INITIALIZATION
    MAINTENANCE = uavcan.node.Mode_1_0.MAINTENANCE
    SOFTWARE_UPDATE = uavcan.node.Mode_1_0.SOFTWARE_UPDATE


VENDOR_SPECIFIC_STATUS_CODE_MASK = (
    2 ** pyuavcan.dsdl.get_model(Heartbeat)["vendor_specific_status_code"].data_type.bit_length_set.max - 1
)


_logger = logging.getLogger(__name__)


class HeartbeatPublisher:
    """
    This class manages periodic publication of the node heartbeat message.
    Also it subscribes to heartbeat messages from other nodes and logs cautionary messages
    if a node-ID conflict is detected on the bus.

    The default states are as follows:

    - Health is NOMINAL.
    - Mode is OPERATIONAL.
    - Vendor-specific status code is zero.
    - Period is MAX_PUBLICATION_PERIOD (see the DSDL definition).
    - Priority is default (i.e., NOMINAL).
    """

    def __init__(self, node: pyuavcan.application.Node):
        self._node = node
        self._health = Health.NOMINAL
        self._mode = Mode.OPERATIONAL
        self._vendor_specific_status_code = 0
        self._pre_heartbeat_handlers: typing.List[typing.Callable[[], None]] = []
        self._maybe_task: typing.Optional[asyncio.Task[None]] = None
        self._priority = pyuavcan.presentation.DEFAULT_PRIORITY
        self._period = float(Heartbeat.MAX_PUBLICATION_PERIOD)
        self._subscriber = self._node.make_subscriber(Heartbeat)
        self._started_at = time.monotonic()

        def start() -> None:
            if not self._maybe_task:
                self._started_at = time.monotonic()
                self._subscriber.receive_in_background(self._handle_received_heartbeat)
                self._maybe_task = self.node.loop.create_task(self._task_function())

        def close() -> None:
            if self._maybe_task:
                self._maybe_task.cancel()  # Cancel first to avoid exceptions from being logged from the task.
                self._maybe_task = None
                self._subscriber.close()

        node.add_lifetime_hooks(start, close)

    @property
    def node(self) -> pyuavcan.application.Node:
        return self._node

    @property
    def uptime(self) -> float:
        """The current amount of time, in seconds, elapsed since the object was instantiated."""
        out = time.monotonic() - self._started_at
        assert out >= 0
        return out

    @property
    def health(self) -> Health:
        """The health value to report with Heartbeat; see :class:`Health`."""
        return self._health

    @health.setter
    def health(self, value: typing.Union[Health, int]) -> None:
        self._health = Health(value)

    @property
    def mode(self) -> Mode:
        """The mode value to report with Heartbeat; see :class:`Mode`."""
        return self._mode

    @mode.setter
    def mode(self, value: typing.Union[Mode, int]) -> None:
        self._mode = Mode(value)

    @property
    def vendor_specific_status_code(self) -> int:
        """The vendor-specific status code (VSSC) value to report with Heartbeat."""
        return self._vendor_specific_status_code

    @vendor_specific_status_code.setter
    def vendor_specific_status_code(self, value: int) -> None:
        value = int(value)
        if 0 <= value <= VENDOR_SPECIFIC_STATUS_CODE_MASK:
            self._vendor_specific_status_code = value
        else:
            raise ValueError(f"Invalid vendor-specific status code: {value}")

    @property
    def period(self) -> float:
        """
        How often the Heartbeat messages should be published. The upper limit (i.e., the lowest frequency)
        is constrained by the UAVCAN specification; please see the DSDL source of ``uavcan.node.Heartbeat``.
        """
        return self._period

    @period.setter
    def period(self, value: float) -> None:
        value = float(value)
        if 0 < value <= Heartbeat.MAX_PUBLICATION_PERIOD:
            self._period = value
        else:
            raise ValueError(f"Invalid heartbeat period: {value}")

    @property
    def priority(self) -> pyuavcan.transport.Priority:
        """
        The transfer priority level to use when publishing Heartbeat messages.
        """
        return self._priority

    @priority.setter
    def priority(self, value: pyuavcan.transport.Priority) -> None:
        # noinspection PyArgumentList
        self._priority = pyuavcan.transport.Priority(value)

    def add_pre_heartbeat_handler(self, handler: typing.Callable[[], None]) -> None:
        """
        Adds a new handler to be invoked immediately before a heartbeat message is published.
        The number of such handlers is unlimited.
        The handler invocation order follows the order of their registration.
        Handlers are invoked from a task running on the node's event loop.
        Handlers are not invoked until the instance is started.

        The handler can be used to synchronize the heartbeat message data (health, mode, vendor-specific status code)
        with external states. Observe that the handler will be invoked even if the heartbeat is not to be published,
        e.g., if the node is anonymous (does not have a node ID). If the handler throws an exception, it will be
        suppressed and logged. Note that the handler is to be not a coroutine but a regular function.

        This is a good method of scheduling periodic status checks on the node.
        """
        self._pre_heartbeat_handlers.append(handler)

    def make_message(self) -> Heartbeat:
        """Constructs a new heartbeat message from the object's state."""
        return Heartbeat(
            uptime=int(self.uptime),  # must floor
            health=uavcan.node.Health_1_0(self.health),
            mode=uavcan.node.Mode_1_0(self.mode),
            vendor_specific_status_code=self.vendor_specific_status_code,
        )

    async def _task_function(self) -> None:
        next_heartbeat_at = time.monotonic()
        pub: typing.Optional[pyuavcan.presentation.Publisher[Heartbeat]] = None
        try:
            while self._maybe_task:
                try:
                    pyuavcan.util.broadcast(self._pre_heartbeat_handlers)()
                    if self.node.id is not None:
                        if pub is None:
                            pub = self.node.make_publisher(Heartbeat)
                        assert pub is not None
                        pub.priority = self._priority
                        if not await pub.publish(self.make_message()):
                            _logger.warning("%s heartbeat send timed out", self)
                except Exception as ex:  # pragma: no cover
                    if (
                        isinstance(ex, (asyncio.CancelledError, pyuavcan.transport.ResourceClosedError))
                        or not self._maybe_task
                    ):
                        _logger.debug("%s publisher task will exit: %s", self, ex)
                        break
                    _logger.exception("%s publisher task exception: %s", self, ex)

                next_heartbeat_at += self._period
                await asyncio.sleep(next_heartbeat_at - time.monotonic())
        finally:
            _logger.debug("%s publisher task is stopping", self)
            if pub is not None:
                pub.close()

    async def _handle_received_heartbeat(self, msg: Heartbeat, metadata: pyuavcan.transport.TransferFrom) -> None:
        local_node_id = self.node.id
        remote_node_id = metadata.source_node_id
        if local_node_id is not None and remote_node_id is not None and local_node_id == remote_node_id:
            _logger.info(
                "NODE-ID CONFLICT: There is another node on the network that uses the same node-ID %d. "
                "Its latest heartbeat is %s with transfer metadata %s",
                remote_node_id,
                msg,
                metadata,
            )

    def __repr__(self) -> str:
        return pyuavcan.util.repr_attributes(
            self,
            heartbeat=self.make_message(),
            priority=self._priority.name,
            period=self._period,
        )



# Copyright (c) 2019 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

import typing
import ipaddress
from pyuavcan.transport import MessageDataSpecifier, ServiceDataSpecifier

IPAddress = typing.Union[ipaddress.IPv4Address, ipaddress.IPv6Address]
"""
I wonder why the common base class of IPv4Address and IPv6Address is not public?
"""

IP_ADDRESS_NODE_ID_MASK = 0xFFFF
"""
Masks the least significant bits of the IP address (v4/v6) that represent the node-ID.
Does not apply to multicast group address.
"""

MULTICAST_GROUP_SUBJECT_ID_MASK = 0xFFFF
"""
Masks the least significant bits of the multicast group address (v4/v6) that represent the subject-ID.
"""

SUBJECT_PORT = 16383
"""
All subjects use the same fixed destination UDP port number.
Subjects are differentiated by the IP multicast group address instead.
"""

SERVICE_BASE_PORT = 16384
"""
Service transfers use unicast IP packets.
Services are differentiated by the destination UDP port number.
See :func:`get_service_udp_port`.
"""


def node_id_to_unicast_ip(local_ip_address: IPAddress, node_id: int) -> IPAddress:
    """
    The local IP address is needed to deduce the subnet that the UAVCAN/UDP transport is operating on.
    The function simply combines the most significant bits from the first argument with the second argument.

    >>> from ipaddress import ip_address
    >>> str(node_id_to_unicast_ip(ip_address('127.42.11.22'), 123))
    '127.42.0.123'
    >>> str(node_id_to_unicast_ip(ip_address('127.42.11.22'), 456))
    '127.42.1.200'
    >>> str(node_id_to_unicast_ip(ip_address('239.42.11.22'), 456))
    Traceback (most recent call last):
      ...
    ValueError: The local address shall be a unicast address, not multicast: 239.42.11.22
    >>> str(node_id_to_unicast_ip(ip_address('127.42.11.22'), 65536))
    Traceback (most recent call last):
      ...
    ValueError: Invalid node-ID...
    """
    if node_id > IP_ADDRESS_NODE_ID_MASK:
        raise ValueError(f"Invalid node-ID: {node_id} is larger than {IP_ADDRESS_NODE_ID_MASK}")
    ty: type
    if isinstance(local_ip_address, ipaddress.IPv4Address):
        mask = 2 ** ipaddress.IPV4LENGTH - 1
        ty = ipaddress.IPv4Address
    elif isinstance(local_ip_address, ipaddress.IPv6Address):
        mask = 2 ** ipaddress.IPV6LENGTH - 1
        ty = ipaddress.IPv6Address
    else:
        assert False
    if local_ip_address.is_multicast:
        raise ValueError(f"The local address shall be a unicast address, not multicast: {local_ip_address}")
    return ty((int(local_ip_address) & (mask ^ IP_ADDRESS_NODE_ID_MASK)) | node_id)  # type: ignore


def unicast_ip_to_node_id(local_ip_address: IPAddress, node_ip_address: IPAddress) -> typing.Optional[int]:
    """
    Returns the node-ID if the node IP address and the local IP address belong to the same subnet.
    Returns None if the node is not a member of the local subnet.
    Raises a value error if either address is a multicast group address.

    >>> from ipaddress import ip_address
    >>> unicast_ip_to_node_id(ip_address('127.42.1.1'), ip_address('127.42.1.200'))
    456
    >>> unicast_ip_to_node_id(ip_address('127.0.0.99'), ip_address('127.0.0.99'))
    99
    >>> unicast_ip_to_node_id(ip_address('127.99.1.1'), ip_address('127.42.1.200'))  # Returns None
    >>> unicast_ip_to_node_id(ip_address('239.42.1.1'), ip_address('127.42.1.200'))
    Traceback (most recent call last):
      ...
    ValueError: Multicast group address cannot be a local IP address...
    >>> unicast_ip_to_node_id(ip_address('127.42.1.1'), ip_address('239.42.1.200'))
    Traceback (most recent call last):
      ...
    ValueError: Multicast group address cannot be mapped to a node-ID...
    """
    if local_ip_address.is_multicast:
        raise ValueError(f"Multicast group address cannot be a local IP address: {local_ip_address}")
    if node_ip_address.is_multicast:
        raise ValueError(f"Multicast group address cannot be mapped to a node-ID: {node_ip_address}")
    if (int(local_ip_address) | IP_ADDRESS_NODE_ID_MASK) == (int(node_ip_address) | IP_ADDRESS_NODE_ID_MASK):
        return int(node_ip_address) & IP_ADDRESS_NODE_ID_MASK
    return None


def message_data_specifier_to_multicast_group(
    local_ip_address: IPAddress, data_specifier: MessageDataSpecifier
) -> IPAddress:
    r"""
    The local IP address is needed to deduce the subnet that the UAVCAN/UDP transport is operating on.
    For IPv4, the resulting address is constructed as follows::

       11101111.0ddddddd.ssssssss.ssssssss
                 \_____/ \_______________/
                subnet-ID    subject-ID

    Where the subnet-ID is taken from the local IP address::

       xxxxxxxx.xddddddd.nnnnnnnn.nnnnnnnn
                 \_____/
                subnet-ID

    >>> from pyuavcan.transport import MessageDataSpecifier
    >>> from ipaddress import ip_address
    >>> str(message_data_specifier_to_multicast_group(ip_address('127.42.11.22'), MessageDataSpecifier(123)))
    '239.42.0.123'
    >>> str(message_data_specifier_to_multicast_group(ip_address('192.168.11.22'), MessageDataSpecifier(456)))
    '239.40.1.200'
    >>> str(message_data_specifier_to_multicast_group(ip_address('239.168.11.22'), MessageDataSpecifier(456)))
    Traceback (most recent call last):
      ...
    ValueError: The local address shall be a unicast address, not multicast: 239.168.11.22
    """
    assert data_specifier.subject_id <= MULTICAST_GROUP_SUBJECT_ID_MASK, "Protocol design error"
    ty: type
    if isinstance(local_ip_address, ipaddress.IPv4Address):
        ty = ipaddress.IPv4Address
        fix = 0b_11101111_00000000_00000000_00000000
        sub = 0b_00000000_01111111_00000000_00000000 & int(local_ip_address)
        msb = fix | sub
    elif isinstance(local_ip_address, ipaddress.IPv6Address):
        raise NotImplementedError("IPv6 is not yet supported; please, submit patches!")
    else:
        assert False
    if local_ip_address.is_multicast:
        raise ValueError(f"The local address shall be a unicast address, not multicast: {local_ip_address}")
    return ty(msb | data_specifier.subject_id)  # type: ignore


def multicast_group_to_message_data_specifier(
    local_ip_address: IPAddress, multicast_group: IPAddress
) -> typing.Optional[MessageDataSpecifier]:
    """
    The inverse of :func:`message_data_specifier_to_multicast_group`.
    The local IP address is needed to ensure that the multicast group belongs to the correct UAVCAN/UDP subnet.
    The return value is None if the multicast group is not valid per the current UAVCAN/UDP specification
    or if it belongs to a different UAVCAN/UDP subnet.

    >>> from ipaddress import ip_address
    >>> multicast_group_to_message_data_specifier(ip_address('127.42.11.22'), ip_address('239.42.1.200'))
    MessageDataSpecifier(subject_id=456)
    >>> multicast_group_to_message_data_specifier(ip_address('127.42.11.22'), ip_address('239.43.1.200'))    # -> None
    >>> multicast_group_to_message_data_specifier(ip_address('127.42.11.22'), ip_address('239.42.255.200'))  # -> None
    """
    try:
        candidate = MessageDataSpecifier(int(multicast_group) & MULTICAST_GROUP_SUBJECT_ID_MASK)
    except ValueError:
        return None
    if message_data_specifier_to_multicast_group(local_ip_address, candidate) == multicast_group:
        return candidate
    return None


def service_data_specifier_to_udp_port(ds: ServiceDataSpecifier) -> int:
    """
    For request transfers, the destination port is computed as
    :data:`SERVICE_BASE_PORT` plus service-ID multiplied by two.
    For response transfers, it is as above plus one.

    >>> service_data_specifier_to_udp_port(ServiceDataSpecifier(0, ServiceDataSpecifier.Role.REQUEST))
    16384
    >>> service_data_specifier_to_udp_port(ServiceDataSpecifier(0, ServiceDataSpecifier.Role.RESPONSE))
    16385
    >>> service_data_specifier_to_udp_port(ServiceDataSpecifier(511, ServiceDataSpecifier.Role.REQUEST))
    17406
    >>> service_data_specifier_to_udp_port(ServiceDataSpecifier(511, ServiceDataSpecifier.Role.RESPONSE))
    17407
    """
    request = SERVICE_BASE_PORT + ds.service_id * 2
    if ds.role == ServiceDataSpecifier.Role.REQUEST:
        return request
    if ds.role == ServiceDataSpecifier.Role.RESPONSE:
        return request + 1
    assert False


def udp_port_to_service_data_specifier(port: int) -> typing.Optional[ServiceDataSpecifier]:
    """
    The inverse of :func:`service_data_specifier_to_udp_port`. Returns None for invalid ports.

    >>> udp_port_to_service_data_specifier(16384)
    ServiceDataSpecifier(service_id=0, role=...REQUEST...)
    >>> udp_port_to_service_data_specifier(16385)
    ServiceDataSpecifier(service_id=0, role=...RESPONSE...)
    >>> udp_port_to_service_data_specifier(17406)
    ServiceDataSpecifier(service_id=511, role=...REQUEST...)
    >>> udp_port_to_service_data_specifier(17407)
    ServiceDataSpecifier(service_id=511, role=...RESPONSE...)
    >>> udp_port_to_service_data_specifier(50000)  # Returns None
    >>> udp_port_to_service_data_specifier(10000)  # Returns None
    """
    out: typing.Optional[ServiceDataSpecifier] = None
    try:
        if port >= SERVICE_BASE_PORT:
            role = ServiceDataSpecifier.Role.REQUEST if port % 2 == 0 else ServiceDataSpecifier.Role.RESPONSE
            out = ServiceDataSpecifier((port - SERVICE_BASE_PORT) // 2, role)
    except ValueError:
        out = None
    return out



# Copyright (c) 2019 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

from __future__ import annotations
import abc
import sys
import typing
import struct

import numpy

_Byte = numpy.uint8
"""
We must use uint8 instead of ubyte because uint8 is platform-invariant whereas (u)byte is platform-dependent.
"""

_EXTRA_BUFFER_CAPACITY_BYTES = 1
"""
We extend the requested buffer size by one because some of the non-byte-aligned write operations
require us to temporarily use one extra byte after the current byte.
"""


class Serializer(abc.ABC):
    """
    All methods operating on scalars implicitly truncate the value if it exceeds the range,
    excepting signed integers, for which overflow handling is not implemented (DSDL does not permit truncation
    of signed integers anyway so it doesn't matter). Saturation must be implemented externally.
    Methods that expect an unsigned integer will raise ValueError if the supplied integer is negative.
    """

    def __init__(self, buffer: numpy.ndarray):
        """
        Do not call this directly. Use :meth:`new` to instantiate.
        """
        self._buf = buffer
        self._bit_offset = 0

    @staticmethod
    def new(buffer_size_in_bytes: int) -> Serializer:
        buffer_size_in_bytes = int(buffer_size_in_bytes) + _EXTRA_BUFFER_CAPACITY_BYTES
        buf: numpy.ndarray = numpy.zeros(buffer_size_in_bytes, dtype=_Byte)
        return _PlatformSpecificSerializer(buf)

    @property
    def current_bit_length(self) -> int:
        return self._bit_offset

    @property
    def buffer(self) -> numpy.ndarray:
        """Returns a properly sized read-only slice of the destination buffer zero-bit-padded to byte."""
        out = self._buf[: (self._bit_offset + 7) // 8]
        out.flags.writeable = False
        # Here we used to check if out.base is self._buf to make sure we're not creating a copy because that might
        # be costly. We no longer do that because it doesn't work with forked serializers: forks don't own their
        # buffers so this check would be failing; also, with MyPy v1.19 this expression used to segfault the
        # interpreter. Very dangerous.
        return out

    def skip_bits(self, bit_length: int) -> None:
        """This is used for padding bits and for skipping fragments written by forked serializers."""
        self._bit_offset += bit_length

    def pad_to_alignment(self, bit_length: int) -> None:
        while self._bit_offset % bit_length != 0:
            self.add_unaligned_bit(False)

    def fork_bytes(self, forked_buffer_size_in_bytes: int) -> Serializer:
        """
        Creates another serializer that uses the same underlying serialization destination buffer
        but offset by :prop:`current_bit_length`. This is intended for delimited serialization.
        The algorithm is simple:

        - Fork the main serializer (M) at the point where the delimited nested instance needs to be serialized.
        - Having obtained the forked serializer (F), skip the size of the delimited header and serialize the object.
        - Take the offset of F (in bytes) sans the size of the delimiter header and serialize the value using M.
        - Skip M by the above number of bytes to avoid overwriting the fragment written by F.
        - Discard F. The job is done.

        This may be unnecessary if the nested object is of a fixed size. In this case, since its length is known,
        the delimiter header can be serialized as a constant, and then the nested object can be serialized trivially
        as if it was sealed.

        This method raises a :class:`ValueError` if the forked instance is not byte-aligned or if the requested buffer
        size is too large.
        """
        if self._bit_offset % 8 != 0:
            raise ValueError("Cannot fork unaligned serializer")
        forked_buffer = self._buf[self._bit_offset // 8 :]
        forked_buffer_size_in_bytes += _EXTRA_BUFFER_CAPACITY_BYTES
        if len(forked_buffer) < forked_buffer_size_in_bytes:
            raise ValueError(
                f"The required forked buffer size of {forked_buffer_size_in_bytes} bytes is less "
                f"than the available remaining buffer space of {len(forked_buffer)} bytes"
            )
        forked_buffer = forked_buffer[:forked_buffer_size_in_bytes]
        assert len(forked_buffer) == forked_buffer_size_in_bytes
        return _PlatformSpecificSerializer(forked_buffer)

    #
    # Fast methods optimized for aligned primitive fields.
    # The most specialized methods must be used whenever possible for best performance.
    #
    @abc.abstractmethod
    def add_aligned_array_of_standard_bit_length_primitives(self, x: numpy.ndarray) -> None:
        """
        Accepts an array of ``(u?int|float)(8|16|32|64)`` and encodes it into the destination.
        On little-endian platforms this may be implemented virtually through ``memcpy()``.
        The current bit offset must be byte-aligned.
        """
        raise NotImplementedError

    def add_aligned_array_of_bits(self, x: numpy.ndarray) -> None:
        """
        Accepts an array of bools and encodes it into the destination using fast native serialization routine
        implemented in numpy. The current bit offset must be byte-aligned.
        """
        assert x.dtype in (bool, numpy.bool_)
        assert self._bit_offset % 8 == 0
        packed = numpy.packbits(x, bitorder="little")
        assert len(packed) * 8 >= len(x)
        self._buf[self._byte_offset : self._byte_offset + len(packed)] = packed
        self._bit_offset += len(x)

    def add_aligned_bytes(self, x: numpy.ndarray) -> None:
        """Simply adds a sequence of bytes; the current bit offset must be byte-aligned."""
        assert self._bit_offset % 8 == 0
        assert x.dtype == _Byte
        self._buf[self._byte_offset : self._byte_offset + len(x)] = x
        self._bit_offset += len(x) * 8

    def add_aligned_u8(self, x: int) -> None:
        assert self._bit_offset % 8 == 0
        self._ensure_not_negative(x)
        self._buf[self._byte_offset] = x
        self._bit_offset += 8

    def add_aligned_u16(self, x: int) -> None:
        self._ensure_not_negative(x)
        self.add_aligned_u8(x & 0xFF)
        self.add_aligned_u8((x >> 8) & 0xFF)

    def add_aligned_u32(self, x: int) -> None:
        self.add_aligned_u16(x)
        self.add_aligned_u16(x >> 16)

    def add_aligned_u64(self, x: int) -> None:
        self.add_aligned_u32(x)
        self.add_aligned_u32(x >> 32)

    def add_aligned_i8(self, x: int) -> None:
        self.add_aligned_u8((256 + x) if x < 0 else x)

    def add_aligned_i16(self, x: int) -> None:
        self.add_aligned_u16((65536 + x) if x < 0 else x)

    def add_aligned_i32(self, x: int) -> None:
        self.add_aligned_u32((2 ** 32 + x) if x < 0 else x)

    def add_aligned_i64(self, x: int) -> None:
        self.add_aligned_u64((2 ** 64 + x) if x < 0 else x)

    def add_aligned_f16(self, x: float) -> None:
        self.add_aligned_bytes(self._float_to_bytes("e", x))

    def add_aligned_f32(self, x: float) -> None:
        self.add_aligned_bytes(self._float_to_bytes("f", x))

    def add_aligned_f64(self, x: float) -> None:
        self.add_aligned_bytes(self._float_to_bytes("d", x))

    #
    # Less specialized methods: assuming that the value is aligned at the beginning, but its bit length
    # is non-standard and may not be an integer multiple of eight.
    # These must not be used if there is a suitable more specialized version defined above.
    #
    def add_aligned_unsigned(self, value: int, bit_length: int) -> None:
        assert self._bit_offset % 8 == 0
        self._ensure_not_negative(value)
        bs = self._unsigned_to_bytes(value, bit_length)
        self._buf[self._byte_offset : self._byte_offset + len(bs)] = bs
        self._bit_offset += bit_length

    def add_aligned_signed(self, value: int, bit_length: int) -> None:
        assert bit_length >= 2
        self.add_aligned_unsigned((2 ** bit_length + value) if value < 0 else value, bit_length)

    #
    # Least specialized methods: no assumptions about alignment are made.
    # These are the slowest and may be used only if none of the above (specialized) methods are suitable.
    #
    @abc.abstractmethod
    def add_unaligned_array_of_standard_bit_length_primitives(self, x: numpy.ndarray) -> None:
        """See the aligned counterpart."""
        raise NotImplementedError

    def add_unaligned_array_of_bits(self, x: numpy.ndarray) -> None:
        assert x.dtype in (bool, numpy.bool_)
        packed = numpy.packbits(x, bitorder="little")
        backtrack = len(packed) * 8 - len(x)
        assert backtrack >= 0
        self.add_unaligned_bytes(packed)
        self._bit_offset -= backtrack

    def add_unaligned_bytes(self, value: numpy.ndarray) -> None:
        assert value.dtype == _Byte
        # This is a faster variant of Ben Dyer's unaligned bit copy algorithm:
        # https://github.com/UAVCAN/libuavcan/blob/fd8ba19bc9c09c05a/libuavcan/src/marshal/uc_bit_array_copy.cpp#L12
        # It is faster because here we are aware that the source is always aligned, which we take advantage of.
        left = self._bit_offset % 8
        right = 8 - left
        for b in value:
            self._buf[self._byte_offset] |= (b << left) & 0xFF
            self._bit_offset += 8
            self._buf[self._byte_offset] = b >> right

    def add_unaligned_unsigned(self, value: int, bit_length: int) -> None:
        self._ensure_not_negative(value)
        bs = self._unsigned_to_bytes(value, bit_length)
        backtrack = len(bs) * 8 - bit_length
        assert backtrack >= 0
        self.add_unaligned_bytes(bs)
        self._bit_offset -= backtrack

    def add_unaligned_signed(self, value: int, bit_length: int) -> None:
        assert bit_length >= 2
        self.add_unaligned_unsigned((2 ** bit_length + value) if value < 0 else value, bit_length)

    def add_unaligned_f16(self, x: float) -> None:
        self.add_unaligned_bytes(self._float_to_bytes("e", x))

    def add_unaligned_f32(self, x: float) -> None:
        self.add_unaligned_bytes(self._float_to_bytes("f", x))

    def add_unaligned_f64(self, x: float) -> None:
        self.add_unaligned_bytes(self._float_to_bytes("d", x))

    def add_unaligned_bit(self, x: bool) -> None:
        self._buf[self._byte_offset] |= bool(x) << (self._bit_offset % 8)
        self._bit_offset += 1

    #
    # Private methods.
    #
    @staticmethod
    def _unsigned_to_bytes(value: int, bit_length: int) -> numpy.ndarray:
        assert bit_length >= 1
        assert value >= 0, "This operation is undefined for negative integers"
        value &= 2 ** bit_length - 1
        num_bytes = (bit_length + 7) // 8
        out = numpy.zeros(num_bytes, dtype=_Byte)
        for i in range(num_bytes):  # Oh, why is my life like this?
            out[i] = value & 0xFF
            value >>= 8
        return out

    @staticmethod
    def _float_to_bytes(format_char: str, x: float) -> numpy.ndarray:
        f = "<" + format_char
        try:
            out = struct.pack(f, x)
        except OverflowError:  # Oops, let's truncate (saturation must be implemented by the caller if needed)
            out = struct.pack(f, numpy.inf if x > 0 else -numpy.inf)
        return numpy.frombuffer(out, dtype=_Byte)  # Note: this operation does not copy the underlying bytes

    @staticmethod
    def _ensure_not_negative(x: int) -> None:
        if x < 0:
            raise ValueError(f"The requested serialization method is not defined on negative integers ({x})")

    @property
    def _byte_offset(self) -> int:
        return self._bit_offset // 8

    def __str__(self) -> str:
        s = " ".join(map(_byte_as_bit_string, self.buffer))
        if self._bit_offset % 8 != 0:
            s, tail = s.rsplit(maxsplit=1)
            bits_to_cut_off = 8 - self._bit_offset % 8
            tail = ("x" * bits_to_cut_off) + tail[bits_to_cut_off:]
            return s + " " + tail
        return s

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self})"


class _LittleEndianSerializer(Serializer):
    # noinspection PyUnresolvedReferences
    def add_aligned_array_of_standard_bit_length_primitives(self, x: numpy.ndarray) -> None:
        # This is close to direct memcpy() from the source memory into the destination memory, which is very fast.
        # We assume that the local platform uses IEEE 754-compliant floating point representation; otherwise,
        # the generated serialized representation may be incorrect. NumPy seems to only support IEEE-754 compliant
        # platforms though so I don't expect any compatibility issues.
        assert x.dtype not in (bool, numpy.bool_, object)
        self.add_aligned_bytes(x.view(_Byte))

    def add_unaligned_array_of_standard_bit_length_primitives(self, x: numpy.ndarray) -> None:
        # This is much slower than the aligned version because we have to manually copy and shift each byte,
        # but still better than manual elementwise serialization.
        assert x.dtype not in (bool, numpy.bool_, object)
        self.add_unaligned_bytes(x.view(_Byte))


class _BigEndianSerializer(Serializer):
    def add_aligned_array_of_standard_bit_length_primitives(self, x: numpy.ndarray) -> None:  # pragma: no cover
        raise NotImplementedError("Pull requests are welcome")

    def add_unaligned_array_of_standard_bit_length_primitives(self, x: numpy.ndarray) -> None:
        raise NotImplementedError("Pull requests are welcome")


_PlatformSpecificSerializer = {
    "little": _LittleEndianSerializer,
    "big": _BigEndianSerializer,
}[sys.byteorder]


def _byte_as_bit_string(x: int) -> str:
    return bin(x)[2:].zfill(8)


def _unittest_serializer_to_str() -> None:
    ser = Serializer.new(50)
    assert str(ser) == ""
    ser.add_aligned_u8(0b11001110)
    assert str(ser) == "11001110"
    ser.add_aligned_i16(-1)
    assert str(ser) == "11001110 11111111 11111111"
    ser.add_aligned_unsigned(0, 1)
    assert str(ser) == "11001110 11111111 11111111 xxxxxxx0"
    ser.add_unaligned_signed(-1, 3)
    assert str(ser) == "11001110 11111111 11111111 xxxx1110"


def _unittest_serializer_aligned() -> None:
    from pytest import raises

    def unseparate(s: typing.Any) -> str:
        return str(s).replace(" ", "")

    bs = _byte_as_bit_string
    ser = Serializer.new(50)
    expected = ""
    assert str(ser) == ""

    with raises(ValueError):
        ser.add_aligned_u8(-42)

    ser.add_aligned_u8(0b1010_0111)
    expected += "1010 0111"
    assert unseparate(ser) == unseparate(expected)

    ser.add_aligned_i64(0x1234_5678_90AB_CDEF)
    expected += bs(0xEF) + bs(0xCD) + bs(0xAB) + bs(0x90)
    expected += bs(0x78) + bs(0x56) + bs(0x34) + bs(0x12)
    assert unseparate(ser) == unseparate(expected)

    ser.add_aligned_i32(-0x1234_5678)  # Two's complement: 0xedcb_a988
    expected += bs(0x88) + bs(0xA9) + bs(0xCB) + bs(0xED)
    assert unseparate(ser) == unseparate(expected)

    ser.add_aligned_i16(-2)  # Two's complement: 0xfffe
    ser.skip_bits(8)
    ser.add_aligned_i8(127)
    expected += bs(0xFE) + bs(0xFF) + bs(0x00) + bs(0x7F)
    assert unseparate(ser) == unseparate(expected)

    ser.add_aligned_f64(1)  # IEEE 754: 0x3ff0_0000_0000_0000
    expected += bs(0x00) * 6 + bs(0xF0) + bs(0x3F)
    ser.add_aligned_f32(1)  # IEEE 754: 0x3f80_0000
    expected += bs(0x00) * 2 + bs(0x80) + bs(0x3F)
    ser.add_aligned_f16(99999.9)  # IEEE 754: overflow, degenerates to +inf: 0x7c00
    expected += bs(0x00) * 1 + bs(0x7C)
    assert unseparate(ser) == unseparate(expected)

    ser.add_aligned_unsigned(0xBEDA, 12)  # 0xBxxx will be truncated away
    expected += "1101 1010 xxxx1110"
    assert unseparate(ser) == unseparate(expected)

    ser.skip_bits(4)  # Bring back into alignment
    expected = expected[:-8] + "00001110"
    assert unseparate(ser) == unseparate(expected)

    ser.add_aligned_unsigned(0xBEDA, 16)  # Making sure byte-size-aligned are handled well, too
    expected += bs(0xDA) + bs(0xBE)
    assert unseparate(ser) == unseparate(expected)

    ser.add_aligned_signed(-2, 9)  # Two's complement: 510 = 0b1_1111_1110
    expected += "11111110 xxxxxxx1"  # MSB is at the end
    assert unseparate(ser) == unseparate(expected)

    ser.skip_bits(7)  # Bring back into alignment
    expected = expected[:-8] + "00000001"
    assert unseparate(ser) == unseparate(expected)

    ser.add_aligned_array_of_standard_bit_length_primitives(numpy.array([0xDEAD, 0xBEEF], numpy.uint16))
    expected += bs(0xAD) + bs(0xDE) + bs(0xEF) + bs(0xBE)
    assert unseparate(ser) == unseparate(expected)

    ser.add_aligned_array_of_bits(
        numpy.array(
            [
                True,
                False,
                True,
                False,
                False,
                False,
                True,
                True,  # 10100011
                True,
                True,
                True,
                False,
                False,
                True,
                True,
                False,  # 11100110
            ],
            bool,
        )
    )
    expected += "11000101 01100111"
    assert unseparate(ser) == unseparate(expected)

    ser.add_aligned_array_of_bits(
        numpy.array(
            [
                True,
                False,
                True,
                False,
                False,
                False,
                True,
                True,  # 10100011
                True,
                True,
                False,
                True,
                False,  # 11010
            ],
            bool,
        )
    )
    expected += "11000101 xxx01011"
    assert unseparate(ser) == unseparate(expected)

    print("repr(serializer):", repr(ser))

    with raises(ValueError, match=".*read-only.*"):
        ser.buffer[0] = 123  # The buffer is read-only for safety reasons


def _unittest_serializer_unaligned() -> None:  # Tricky cases with unaligned fields (very tricky)
    ser = Serializer.new(40)

    ser.add_unaligned_array_of_bits(
        numpy.array(
            [
                True,
                False,
                True,
                False,
                False,
                False,
                True,
                True,  # 10100011
                True,
                True,
                True,  # 111
            ],
            bool,
        )
    )
    assert str(ser) == "11000101 xxxxx111"

    ser.add_unaligned_array_of_bits(
        numpy.array(
            [
                True,
                False,
                True,
                False,
                False,  # ???10100 (byte alignment restored here)
                True,
                True,
                True,
                False,
                True,  # 11101 (byte alignment lost, three bits short)
            ],
            bool,
        )
    )
    assert str(ser) == "11000101 00101111 xxx10111"

    # Adding '00010010 00110100 01010110'
    ser.add_unaligned_bytes(numpy.array([0x12, 0x34, 0x56], dtype=_Byte))
    assert str(ser) == "11000101 00101111 01010111 10000010 11000110 xxx01010"

    ser.add_unaligned_array_of_bits(numpy.array([False, True, True], bool))
    assert ser._bit_offset % 8 == 0, "Byte alignment is not restored"  # pylint: disable=protected-access
    assert str(ser) == "11000101 00101111 01010111 10000010 11000110 11001010"

    ser.add_unaligned_bytes(numpy.array([0x12, 0x34, 0x56], dtype=_Byte))  # We're actually aligned here
    assert str(ser) == "11000101 00101111 01010111 10000010 11000110 11001010 00010010 00110100 01010110"

    ser.add_unaligned_bit(True)
    ser.add_unaligned_bit(False)
    ser.add_unaligned_bit(False)
    ser.add_unaligned_bit(True)
    ser.add_unaligned_bit(True)  # Three bits short until alignment
    assert str(ser) == "11000101 00101111 01010111 10000010 11000110 11001010 00010010 00110100 01010110 xxx11001"

    ser.add_unaligned_signed(-2, 8)  # Two's complement: 254 = 1111 1110
    assert (
        str(ser) == "11000101 00101111 01010111 10000010 11000110 11001010 00010010 00110100 01010110 11011001 "
        "xxx11111"
    )

    ser.add_unaligned_unsigned(0b11101100101, 11)  # Tricky, eh? Eleven bits, unaligned write
    assert ser._bit_offset % 8 == 0, "Byte alignment is not restored"  # pylint: disable=protected-access
    assert (
        str(ser) == "11000101 00101111 01010111 10000010 11000110 11001010 00010010 00110100 01010110 11011001 "
        "10111111 11101100"
    )

    ser.add_unaligned_unsigned(0b1110, 3)  # MSB truncated away
    assert (
        str(ser) == "11000101 00101111 01010111 10000010 11000110 11001010 00010010 00110100 01010110 11011001 "
        "10111111 11101100 xxxxx110"
    )

    # Adding '00000000 00000000 00000000 00000000 00000000 00000000 11110000 00111111'
    ser.add_unaligned_f64(1)
    assert (
        str(ser) == "11000101 00101111 01010111 10000010 11000110 11001010 00010010 00110100 01010110 11011001 "
        "10111111 11101100 00000110 00000000 00000000 00000000 00000000 00000000 10000000 11111111 "
        "xxxxx001"
    )

    # Adding '00000000 00000000 10000000 00111111'
    ser.add_unaligned_f32(1)
    assert (
        str(ser) == "11000101 00101111 01010111 10000010 11000110 11001010 00010010 00110100 01010110 11011001 "
        "10111111 11101100 00000110 00000000 00000000 00000000 00000000 00000000 10000000 11111111 "
        "00000001 00000000 00000000 11111100 xxxxx001"
    )

    # Adding '00000000 11111100'
    ser.add_unaligned_f16(-99999.9)
    assert (
        str(ser) == "11000101 00101111 01010111 10000010 11000110 11001010 00010010 00110100 01010110 11011001 "
        "10111111 11101100 00000110 00000000 00000000 00000000 00000000 00000000 10000000 11111111 "
        "00000001 00000000 00000000 11111100 00000001 11100000 xxxxx111"
    )

    # Adding '10101101 11011110 11101111 10111110'
    ser.add_unaligned_array_of_standard_bit_length_primitives(numpy.array([0xDEAD, 0xBEEF], numpy.uint16))
    assert (
        str(ser) == "11000101 00101111 01010111 10000010 11000110 11001010 00010010 00110100 01010110 11011001 "
        "10111111 11101100 00000110 00000000 00000000 00000000 00000000 00000000 10000000 11111111 "
        "00000001 00000000 00000000 11111100 00000001 11100000 01101111 11110101 01111110 11110111 "
        "xxxxx101"
    )

    ser.skip_bits(5)
    assert ser._bit_offset % 8 == 0, "Byte alignment is not restored"  # pylint: disable=protected-access
    assert (
        str(ser) == "11000101 00101111 01010111 10000010 11000110 11001010 00010010 00110100 01010110 11011001 "
        "10111111 11101100 00000110 00000000 00000000 00000000 00000000 00000000 10000000 11111111 "
        "00000001 00000000 00000000 11111100 00000001 11100000 01101111 11110101 01111110 11110111 "
        "00000101"
    )

    print("repr(serializer):", repr(ser))


def _unittest_serializer_fork_bytes() -> None:
    import pytest

    r = Serializer.new(16)
    m = Serializer.new(16)
    assert str(r) == str(m)

    r.add_aligned_u8(123)
    m.add_aligned_u8(123)
    assert str(r) == str(m)

    with pytest.raises(ValueError):
        m.fork_bytes(16)  # Out of range

    f = m.fork_bytes(15)
    assert str(f) == ""
    r.add_aligned_u8(42)
    f.add_aligned_u8(42)
    assert str(r) != str(m)
    m.skip_bits(8)
    assert str(r) == str(m)  # M updated even though we didn't write in it!

    r.add_aligned_u8(11)
    m.add_aligned_u8(11)
    assert str(r) == str(m)

    f.skip_bits(8)
    ff = f.fork_bytes(1)
    r.add_aligned_u8(22)
    ff.add_aligned_u8(22)
    assert str(r) != str(m)
    m.skip_bits(8)
    assert str(r) == str(m)  # M updated even though we didn't write in it! Double indirection.

    ff.add_unaligned_bit(True)  # Break alignment
    with pytest.raises(ValueError):
        ff.fork_bytes(1)  # Bad alignment



# Copyright (c) 2019 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>


def repr_attributes(obj: object, *anonymous_elements: object, **named_elements: object) -> str:
    """
    A simple helper function that constructs a :func:`repr` form of an object. Used widely across the library.
    String representations will be obtained by invoking :func:`str` on each value.

    >>> class Aa: pass
    >>> assert repr_attributes(Aa()) == 'Aa()'
    >>> assert repr_attributes(Aa(), 123) == 'Aa(123)'
    >>> assert repr_attributes(Aa(), foo=123) == 'Aa(foo=123)'
    >>> assert repr_attributes(Aa(), 456, foo=123, bar=repr('abc')) == "Aa(456, foo=123, bar='abc')"
    """
    fld = list(map(str, anonymous_elements)) + list(f"{name}={value}" for name, value in named_elements.items())
    return f"{type(obj).__name__}(" + ", ".join(fld) + ")"


def repr_attributes_noexcept(obj: object, *anonymous_elements: object, **named_elements: object) -> str:
    """
    A robust version of :meth:`repr_attributes` that never raises exceptions.

    >>> class Aa: pass
    >>> repr_attributes_noexcept(Aa(), 456, foo=123, bar=repr('abc'))
    "Aa(456, foo=123, bar='abc')"
    >>> class Bb:
    ...     def __repr__(self) -> str:
    ...         raise Exception('Ford, you are turning into a penguin')
    >>> repr_attributes_noexcept(Aa(), foo=Bb())
    "<REPR FAILED: Exception('Ford, you are turning into a penguin')>"
    >>> class Cc(Exception):
    ...     def __str__(self) -> str:  raise Cc()  # Infinite recursion
    ...     def __repr__(self) -> str: raise Cc()  # Infinite recursion
    >>> repr_attributes_noexcept(Aa(), foo=Cc())
    '<REPR FAILED: UNKNOWN ERROR>'
    """
    try:
        return repr_attributes(obj, *anonymous_elements, **named_elements)
    except Exception as ex:
        # noinspection PyBroadException
        try:
            return f"<REPR FAILED: {ex!r}>"
        except Exception:
            return "<REPR FAILED: UNKNOWN ERROR>"



# Copyright (c) 2019 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

import typing

import numpy
import pydsdl

from ._composite_object import CompositeObject, get_model, get_attribute, set_attribute, get_class
from ._composite_object import CompositeObjectTypeVar


def to_builtin(obj: CompositeObject) -> typing.Dict[str, typing.Any]:
    """
    Accepts a DSDL object (an instance of a Python class auto-generated from a DSDL definition),
    returns its value represented using only native built-in types: dict, list, bool, int, float, str.
    Ordering of dict elements is guaranteed to match the field ordering of the source definition.
    Keys of dicts representing DSDL objects use the original unstropped names from the source DSDL definition;
    e.g., ``if``, not ``if_``.

    This is intended for use with JSON, YAML, and other serialization formats.

    ..  doctest::
        :hide:

        >>> import tests
        >>> _ = tests.dsdl.compile()

    >>> import json
    >>> import uavcan.primitive.array
    >>> json.dumps(to_builtin(uavcan.primitive.array.Integer32_1_0([-123, 456, 0])))
    '{"value": [-123, 456, 0]}'
    >>> import uavcan.register
    >>> request = uavcan.register.Access_1_0.Request(
    ...     uavcan.register.Name_1_0('my.register'),
    ...     uavcan.register.Value_1_0(integer16=uavcan.primitive.array.Integer16_1_0([1, 2, +42, -10_000]))
    ... )
    >>> to_builtin(request)  # doctest: +NORMALIZE_WHITESPACE
    {'name':  {'name': 'my.register'},
     'value': {'integer16': {'value': [1, 2, 42, -10000]}}}
    """
    model = get_model(obj)
    _raise_if_service_type(model)
    out = _to_builtin_impl(obj, model)
    assert isinstance(out, dict)
    return out


def _to_builtin_impl(
    obj: typing.Union[CompositeObject, numpy.ndarray, str, bool, int, float], model: pydsdl.SerializableType
) -> typing.Union[typing.Dict[str, typing.Any], typing.List[typing.Any], str, bool, int, float]:
    if isinstance(model, pydsdl.CompositeType):
        assert isinstance(obj, CompositeObject)
        return {
            f.name: _to_builtin_impl(get_attribute(obj, f.name), f.data_type)
            for f in model.fields_except_padding
            if get_attribute(obj, f.name) is not None  # The check is to hide inactive union variants.
        }

    if isinstance(model, pydsdl.ArrayType):
        assert isinstance(obj, numpy.ndarray)
        if model.string_like:  # TODO: drop this special case when strings are natively supported in DSDL.
            try:
                return bytes(e for e in obj).decode()
            except UnicodeError:
                return list(map(int, obj))
        return [_to_builtin_impl(e, model.element_type) for e in obj]

    if isinstance(model, pydsdl.PrimitiveType):
        # The explicit conversions are needed to get rid of NumPy scalar types.
        if isinstance(model, pydsdl.IntegerType):
            return int(obj)  # type: ignore
        if isinstance(model, pydsdl.FloatType):
            return float(obj)  # type: ignore
        if isinstance(model, pydsdl.BooleanType):
            return bool(obj)
        assert isinstance(obj, str)
        return obj

    assert False, "Unexpected inputs"


def update_from_builtin(destination: CompositeObjectTypeVar, source: typing.Any) -> CompositeObjectTypeVar:
    """
    Updates the provided DSDL object (an instance of a Python class auto-generated from a DSDL definition)
    with the values from a native representation, where DSDL objects are represented as dicts, arrays
    are lists, and primitives are represented as int/float/bool. This is the reverse of :func:`to_builtin`.
    Values that are not specified in the source are not updated (left at their original values),
    so an empty source will leave the input object unchanged.

    Source field names shall match the original unstropped names provided in the DSDL definition;
    e.g., `if`, not `if_`. If there is more than one variant specified for a union type, the last
    specified variant takes precedence.

    :param destination: The object to update. The update will be done in-place. If you don't want the source
        object modified, clone it beforehand.

    :param source: The :class:`dict` instance containing the values to update the destination object with.

    :return: A reference to destination (not a copy).

    :raises: :class:`ValueError` if the provided source values cannot be applied to the destination object,
        or if the source contains fields that are not present in the destination object.
        :class:`TypeError` if an entity of the source cannot be converted into the type expected by the destination.

    >>> import tests; tests.dsdl.compile()  # DSDL package generation not shown in this example.
    [...]
    >>> import json
    >>> import uavcan.primitive.array
    >>> import uavcan.register
    >>> request = uavcan.register.Access_1_0.Request(
    ...     uavcan.register.Name_1_0('my.register'),
    ...     uavcan.register.Value_1_0(string=uavcan.primitive.String_1_0('Hello world!'))
    ... )
    >>> request
    uavcan.register.Access.Request...name='my.register'...value='Hello world!'...
    >>> update_from_builtin(request, {  # Switch the Value union from string to int16; keep the name unchanged.
    ...     'value': {
    ...         'integer16': {
    ...             'value': [1, 2, 42, -10000]
    ...         }
    ...     }
    ... })  # doctest: +NORMALIZE_WHITESPACE
    uavcan.register.Access.Request...name='my.register'...value=[ 1, 2, 42,-10000]...
    """
    # UX improvement; see https://github.com/UAVCAN/pyuavcan/issues/116
    if not isinstance(source, dict):
        raise TypeError(
            f"Invalid format: cannot update an instance of type {type(destination).__name__!r} "
            f"from value of type {type(source).__name__!r}"
        )

    source = dict(source)  # Create copy to prevent mutation of the original

    if not isinstance(destination, CompositeObject):  # pragma: no cover
        raise TypeError(f"Bad destination: expected a CompositeObject, got {type(destination).__name__}")

    model = get_model(destination)
    _raise_if_service_type(model)

    for f in model.fields_except_padding:
        field_type = f.data_type
        try:
            value = source.pop(f.name)
        except LookupError:
            continue  # No value specified, keep original value

        if isinstance(field_type, pydsdl.CompositeType):
            field_obj = get_attribute(destination, f.name)
            if field_obj is None:  # Oh, this is a union
                field_obj = get_class(field_type)()  # The variant was not selected, construct a default
                set_attribute(destination, f.name, field_obj)  # Switch the union to the new variant
            update_from_builtin(field_obj, value)

        elif isinstance(field_type, pydsdl.ArrayType):
            element_type = field_type.element_type
            if isinstance(element_type, pydsdl.PrimitiveType):
                set_attribute(destination, f.name, value)
            elif isinstance(element_type, pydsdl.CompositeType):
                dtype = get_class(element_type)
                set_attribute(destination, f.name, [update_from_builtin(dtype(), s) for s in value])
            else:
                assert False, f"Unexpected array element type: {element_type!r}"

        elif isinstance(field_type, pydsdl.PrimitiveType):
            set_attribute(destination, f.name, value)

        else:
            assert False, f"Unexpected field type: {field_type!r}"

    if source:
        raise ValueError(f"No such fields in {model}: {list(source.keys())}")

    return destination


def _raise_if_service_type(model: pydsdl.SerializableType) -> None:
    if isinstance(model, pydsdl.ServiceType):  # pragma: no cover
        raise TypeError(
            f"Built-in form is not defined for service types. "
            f"Did you mean to use Request or Response? Input type: {model}"
        )



# Copyright (c) 2019 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

import typing
import dataclasses
import pyuavcan.transport
from ._base import Deduplicator


class CyclicDeduplicator(Deduplicator):
    def __init__(self, transfer_id_modulo: int) -> None:
        self._tid_modulo = int(transfer_id_modulo)
        assert self._tid_modulo > 0
        self._remote_states: typing.List[typing.Optional[_RemoteState]] = []

    def should_accept_transfer(
        self,
        *,
        iface_id: int,
        transfer_id_timeout: float,
        timestamp: pyuavcan.transport.Timestamp,
        source_node_id: typing.Optional[int],
        transfer_id: int,
    ) -> bool:
        if source_node_id is None:
            # Anonymous transfers are fully stateless, so always accepted.
            # This may lead to duplications and reordering but this is a design limitation.
            return True

        # If a similar architecture is used on an embedded system, this normally would be a static array.
        if len(self._remote_states) <= source_node_id:
            self._remote_states += [None] * (source_node_id - len(self._remote_states) + 1)
            assert len(self._remote_states) == source_node_id + 1

        if self._remote_states[source_node_id] is None:
            # First transfer from this node, create new state and accept unconditionally.
            self._remote_states[source_node_id] = _RemoteState(iface_id=iface_id, last_timestamp=timestamp)
            return True

        # We have seen transfers from this node before, so we need to perform actual deduplication.
        state = self._remote_states[source_node_id]
        assert state is not None

        # If the current interface was seen working recently, reject traffic from other interfaces.
        # Note that the time delta may be negative due to timestamping variations and inner latency variations.
        time_delta = timestamp.monotonic - state.last_timestamp.monotonic
        iface_switch_allowed = time_delta > transfer_id_timeout
        if not iface_switch_allowed and state.iface_id != iface_id:
            return False

        # TODO: The TID modulo setting is not currently used yet.
        # TODO: It may be utilized later to implement faster iface fallback.

        # Either we're on the same interface or (the interface is new and the current one seems to be down).
        state.iface_id = iface_id
        state.last_timestamp = timestamp
        return True


@dataclasses.dataclass
class _RemoteState:
    iface_id: int
    last_timestamp: pyuavcan.transport.Timestamp



# Copyright (c) 2019 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

from __future__ import annotations
from typing import Callable, Type, TypeVar, Optional, List, Any
import abc
import asyncio
import logging
import uavcan.node
import pyuavcan
from pyuavcan.presentation import Presentation, ServiceRequestMetadata, Publisher, Subscriber, Server, Client
from . import heartbeat_publisher
from . import register


NodeInfo = uavcan.node.GetInfo_1_0.Response

MessageClass = TypeVar("MessageClass", bound=pyuavcan.dsdl.CompositeObject)
ServiceClass = TypeVar("ServiceClass", bound=pyuavcan.dsdl.ServiceObject)


class PortNotConfiguredError(register.MissingRegisterError):
    """
    Raised from :meth:`Node.make_publisher`, :meth:`Node.make_subscriber`, :meth:`Node.make_client`,
    :meth:`Node.get_server` if the application requested a port for which there is no configuration register
    and whose data type does not have a fixed port-ID.

    Applications may catch this exception to implement optional ports,
    where the port is not enabled until explicitly configured while other components of the application are functional.
    """


class Node(abc.ABC):
    """
    This is the top-level abstraction representing a UAVCAN node on the bus.
    This is an abstract class; instantiate it using the factory :func:`pyuavcan.application.make_node`
    or (in special cases) create custom implementations.

    This class automatically instantiates the following application-layer function implementations:

    - :class:`heartbeat_publisher.HeartbeatPublisher`
    - Register API server (``uavcan.register.*``)
    - Node info server (``uavcan.node.GetInfo``)
    - Port introspection publisher (``uavcan.port.List``)

    ..  attention::

        If the underlying transport is anonymous, some of these functions may not be available.

    Start the instance when initialization is finished by invoking :meth:`start`.
    This will also automatically start all function implementation instances.
    """

    def __init__(self) -> None:
        self._started = False
        self._on_start: List[Callable[[], None]] = []
        self._on_close: List[Callable[[], None]] = []

        # Instantiate application-layer functions. Please keep the class docstring updated when changing this.
        self._heartbeat_publisher = heartbeat_publisher.HeartbeatPublisher(self)

        from ._port_list_publisher import PortListPublisher
        from ._register_server import RegisterServer

        PortListPublisher(self)

        async def handle_get_info(_req: uavcan.node.GetInfo_1_0.Request, _meta: ServiceRequestMetadata) -> NodeInfo:
            return self.info

        try:
            RegisterServer(self)
            srv_info = self.get_server(uavcan.node.GetInfo_1_0)
        except pyuavcan.transport.OperationNotDefinedForAnonymousNodeError as ex:
            _logger.info("%r: RPC-servers not launched because the transport is anonymous: %s", self, ex)
        else:
            self.add_lifetime_hooks(lambda: srv_info.serve_in_background(handle_get_info), srv_info.close)

    @property
    @abc.abstractmethod
    def presentation(self) -> Presentation:
        """Provides access to the underlying instance of :class:`pyuavcan.presentation.Presentation`."""
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def info(self) -> NodeInfo:
        """Provides access to the local node info structure. See :class:`pyuavcan.application.NodeInfo`."""
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def registry(self) -> register.Registry:
        """
        Provides access to the local registry instance (see :class:`pyuavcan.application.register.Registry`).
        The registry manages UAVCAN registers as defined by the standard network service ``uavcan.register``.

        The registers store the configuration parameters of the current application, both standard
        (like subject-IDs, service-IDs, transport configuration, the local node-ID, etc.)
        and application-specific ones.

        See also :meth:`make_publisher`, :meth:`make_subscriber`, :meth:`make_client`, :meth:`get_server`.
        """
        raise NotImplementedError

    @property
    def loop(self) -> asyncio.AbstractEventLoop:
        """Shortcut for ``self.presentation.loop``"""
        return self.presentation.loop

    @property
    def id(self) -> Optional[int]:
        """Shortcut for ``self.presentation.transport.local_node_id``"""
        return self.presentation.transport.local_node_id

    @property
    def heartbeat_publisher(self) -> heartbeat_publisher.HeartbeatPublisher:
        """Provides access to the heartbeat publisher instance of this node."""
        return self._heartbeat_publisher

    def make_publisher(self, dtype: Type[MessageClass], port_name: str = "") -> Publisher[MessageClass]:
        """
        Wrapper over :meth:`pyuavcan.presentation.Presentation.make_publisher`
        that takes the subject-ID from the standard register ``uavcan.pub.PORT_NAME.id``.
        If the register is missing or no name is given, the fixed subject-ID is used unless it is also missing.
        The type information is automatically exposed via ``uavcan.pub.PORT_NAME.type`` based on dtype.
        For details on the standard registers see Specification.

        :raises:
            :class:`PortNotConfiguredError` if the register is missing and no fixed port-ID is defined.
            :class:`TypeError` if no name is given and no fixed port-ID is defined.
        """
        if port_name:
            return self.presentation.make_publisher(dtype, self._resolve_named_port(dtype, "pub", port_name))
        return self.presentation.make_publisher_with_fixed_subject_id(dtype)  # type: ignore

    def make_subscriber(self, dtype: Type[MessageClass], port_name: str = "") -> Subscriber[MessageClass]:
        """
        Wrapper over :meth:`pyuavcan.presentation.Presentation.make_subscriber`
        that takes the subject-ID from the standard register ``uavcan.sub.PORT_NAME.id``.
        If the register is missing or no name is given, the fixed subject-ID is used unless it is also missing.
        The type information is automatically exposed via ``uavcan.sub.PORT_NAME.type`` based on dtype.
        For details on the standard registers see Specification.

        :raises:
            :class:`PortNotConfiguredError` if the register is missing and no fixed port-ID is defined.
            :class:`TypeError` if no name is given and no fixed port-ID is defined.
        """
        if port_name:
            return self.presentation.make_subscriber(dtype, self._resolve_named_port(dtype, "sub", port_name))
        return self.presentation.make_subscriber_with_fixed_subject_id(dtype)  # type: ignore

    def make_client(self, dtype: Type[ServiceClass], server_node_id: int, port_name: str = "") -> Client[ServiceClass]:
        """
        Wrapper over :meth:`pyuavcan.presentation.Presentation.make_client`
        that takes the service-ID from the standard register ``uavcan.cln.PORT_NAME.id``.
        If the register is missing or no name is given, the fixed service-ID is used unless it is also missing.
        The type information is automatically exposed via ``uavcan.cln.PORT_NAME.type`` based on dtype.
        For details on the standard registers see Specification.

        :raises:
            :class:`PortNotConfiguredError` if the register is missing and no fixed port-ID is defined.
            :class:`TypeError` if no name is given and no fixed port-ID is defined.
        """
        if port_name:
            return self.presentation.make_client(
                dtype,
                service_id=self._resolve_named_port(dtype, "cln", port_name),
                server_node_id=server_node_id,
            )
        return self.presentation.make_client_with_fixed_service_id(dtype, server_node_id=server_node_id)  # type: ignore

    def get_server(self, dtype: Type[ServiceClass], port_name: str = "") -> Server[ServiceClass]:
        """
        Wrapper over :meth:`pyuavcan.presentation.Presentation.get_server`
        that takes the service-ID from the standard register ``uavcan.srv.PORT_NAME.id``.
        If the register is missing or no name is given, the fixed service-ID is used unless it is also missing.
        The type information is automatically exposed via ``uavcan.srv.PORT_NAME.type`` based on dtype.
        For details on the standard registers see Specification.

        :raises:
            :class:`PortNotConfiguredError` if the register is missing and no fixed port-ID is defined.
            :class:`TypeError` if no name is given and no fixed port-ID is defined.
        """
        if port_name:
            return self.presentation.get_server(dtype, self._resolve_named_port(dtype, "srv", port_name))
        return self.presentation.get_server_with_fixed_service_id(dtype)  # type: ignore

    def _resolve_named_port(self, dtype: Type[pyuavcan.dsdl.CompositeObject], kind: str, name: str) -> int:
        assert name, "Internal error"
        model = pyuavcan.dsdl.get_model(dtype)

        id_register_name = f"uavcan.{kind}.{name}.id"
        port_id = int(
            self.registry.setdefault(
                id_register_name,
                register.Value(natural16=register.Natural16([0xFFFF])),  # 0xFFFF means unset
            )
        )
        # Expose the type information to other network participants as prescribed by the Specification.
        self.registry[f"uavcan.{kind}.{name}.type"] = lambda: register.Value(string=register.String(str(model)))

        # Check if the value stored in the register is actually valid.
        mask = {
            "pub": pyuavcan.transport.MessageDataSpecifier.SUBJECT_ID_MASK,
            "sub": pyuavcan.transport.MessageDataSpecifier.SUBJECT_ID_MASK,
            "cln": pyuavcan.transport.ServiceDataSpecifier.SERVICE_ID_MASK,
            "srv": pyuavcan.transport.ServiceDataSpecifier.SERVICE_ID_MASK,
        }[kind]
        if 0 <= port_id <= mask:
            return port_id

        # Default to the fixed port-ID if the register value is invalid.
        _logger.debug("%r: %r = %r not in [0, %d], assume undefined", self, id_register_name, port_id, mask)
        if model.fixed_port_id is not None:
            assert isinstance(model.fixed_port_id, int)
            return model.fixed_port_id

        raise PortNotConfiguredError(
            id_register_name,
            f"Cannot initialize {kind}-port {name!r} because the register "
            f"does not define a valid port-ID and no fixed port-ID is defined for {model}. "
            f"Check if the environment variables are passed correctly or if the application is using the "
            f"correct register file.",
        )

    def start(self) -> None:
        """
        Starts all application-layer function implementations that are initialized on this node
        (like the heartbeat publisher, diagnostics, and basically anything that takes a node reference
        in its constructor).
        These will be automatically terminated when the node is closed.
        This method is idempotent.
        """
        if not self._started:
            for fun in self._on_start:  # First failure aborts the start.
                fun()
            self._started = True

    def close(self) -> None:
        """
        Closes the :attr:`presentation` (which includes the transport), the registry, the application-layer functions.
        The user does not have to close every port manually as it will be done automatically.
        This method is idempotent.
        Calling :meth:`start` on a closed node may lead to unpredictable results.
        """
        pyuavcan.util.broadcast(self._on_close)()
        self.presentation.close()
        self.registry.close()

    def add_lifetime_hooks(self, start: Optional[Callable[[], None]], close: Optional[Callable[[], None]]) -> None:
        """
        The start hook will be invoked when this node is :meth:`start`-ed.
        If the node is already started when this method is invoked, the start hook is called immediately.

        The close hook is invoked when this node is :meth:`close`-d.
        If the node is already closed, the close hook will never be invoked.
        """
        if start is not None:
            if self._started:
                start()
            else:
                self._on_start.append(start)
        if close is not None:
            self._on_close.append(close)

    def __enter__(self) -> Node:
        """
        Invokes :meth:`start` upon entering the context. Does nothing if already started.
        """
        self.start()
        return self

    def __exit__(self, *_: Any) -> None:
        """
        Invokes :meth:`close` upon leaving the context. Does nothing if already closed.
        """
        self.close()

    def __repr__(self) -> str:
        return pyuavcan.util.repr_attributes(self, self.info, self.presentation, self.registry)


_logger = logging.getLogger(__name__)



# Copyright (c) 2019 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

from __future__ import annotations
import copy
import typing
import asyncio
import logging
import dataclasses
import pyuavcan.util
import pyuavcan.transport
from pyuavcan.transport import Timestamp
from .._frame import UAVCANFrame
from .._identifier import CANID, MessageCANID, ServiceCANID
from ._base import CANSession, SessionFinalizer
from ._transfer_reassembler import TransferReassemblyErrorID, TransferReassembler


_logger = logging.getLogger(__name__)


@dataclasses.dataclass
class CANInputSessionStatistics(pyuavcan.transport.SessionStatistics):
    reception_error_counters: typing.Dict[TransferReassemblyErrorID, int] = dataclasses.field(
        default_factory=lambda: {e: 0 for e in TransferReassemblyErrorID}
    )


class CANInputSession(CANSession, pyuavcan.transport.InputSession):
    DEFAULT_TRANSFER_ID_TIMEOUT = 2
    """
    Per the UAVCAN specification. Units are seconds. Can be overridden after instantiation if needed.
    """

    _QueueItem = typing.Tuple[Timestamp, CANID, UAVCANFrame]

    def __init__(
        self,
        specifier: pyuavcan.transport.InputSessionSpecifier,
        payload_metadata: pyuavcan.transport.PayloadMetadata,
        loop: asyncio.AbstractEventLoop,
        finalizer: SessionFinalizer,
    ):
        """Use the factory method."""
        self._specifier = specifier
        self._payload_metadata = payload_metadata

        self._queue: asyncio.Queue[CANInputSession._QueueItem] = asyncio.Queue()
        assert loop is not None
        self._loop = loop
        self._transfer_id_timeout_ns = int(CANInputSession.DEFAULT_TRANSFER_ID_TIMEOUT / _NANO)

        self._receivers = [TransferReassembler(nid, payload_metadata.extent_bytes) for nid in _node_id_range()]

        self._statistics = CANInputSessionStatistics()  # We could easily support per-source-node statistics if needed

        super().__init__(finalizer=finalizer)

    def _push_frame(self, timestamp: Timestamp, can_id: CANID, frame: UAVCANFrame) -> None:
        """
        This is a part of the transport-internal API. It's a public method despite the name because Python's
        visibility handling capabilities are limited. I guess we could define a private abstract base to
        handle this but it feels like too much work. Why can't we have protected visibility in Python?
        """
        try:
            self._queue.put_nowait((timestamp, can_id, frame))
        except asyncio.QueueFull:
            self._statistics.drops += 1
            _logger.info(
                "%s: Input queue overflow; frame %s (CAN ID fields: %s) received at %s is dropped",
                self,
                frame,
                can_id,
                timestamp,
            )

    @property
    def frame_queue_capacity(self) -> typing.Optional[int]:
        """
        Capacity of the input frame queue. None means that the capacity is unlimited, which is the default.
        This may deplete the heap if input transfers are not consumed quickly enough so beware.

        If the capacity is changed and the new value is smaller than the number of frames currently in the queue,
        the newest frames will be discarded and the number of queue overruns will be incremented accordingly.
        The complexity of a queue capacity change may be up to linear of the number of frames currently in the queue.
        If the value is not None, it must be a positive integer, otherwise you get a :class:`ValueError`.
        """
        return self._queue.maxsize if self._queue.maxsize > 0 else None

    @frame_queue_capacity.setter
    def frame_queue_capacity(self, value: typing.Optional[int]) -> None:
        if value is not None and not value > 0:
            raise ValueError(f"Invalid value for queue capacity: {value}")

        old_queue = self._queue
        self._queue = asyncio.Queue(int(value) if value is not None else 0)
        try:
            while True:
                self._push_frame(*old_queue.get_nowait())
        except asyncio.QueueEmpty:
            pass

    @property
    def specifier(self) -> pyuavcan.transport.InputSessionSpecifier:
        return self._specifier

    @property
    def payload_metadata(self) -> pyuavcan.transport.PayloadMetadata:
        return self._payload_metadata

    def sample_statistics(self) -> CANInputSessionStatistics:
        return copy.copy(self._statistics)

    @property
    def transfer_id_timeout(self) -> float:
        return self._transfer_id_timeout_ns * _NANO

    @transfer_id_timeout.setter
    def transfer_id_timeout(self, value: float) -> None:
        if value > 0:
            self._transfer_id_timeout_ns = round(value / _NANO)
        else:
            raise ValueError(f"Invalid value for transfer-ID timeout [second]: {value}")

    async def receive(self, monotonic_deadline: float) -> typing.Optional[pyuavcan.transport.TransferFrom]:
        out = await self._do_receive(monotonic_deadline)
        assert (
            out is None or self.specifier.remote_node_id is None or out.source_node_id == self.specifier.remote_node_id
        ), "Internal input session protocol violation"
        return out

    def close(self) -> None:  # pylint: disable=useless-super-delegation
        super().close()

    async def _do_receive(self, monotonic_deadline: float) -> typing.Optional[pyuavcan.transport.TransferFrom]:
        while True:
            try:
                # Continue reading past the deadline until the queue is empty or a transfer is received.
                timeout = monotonic_deadline - self._loop.time()
                if timeout > 0:
                    timestamp, canid, frame = await asyncio.wait_for(self._queue.get(), timeout)
                else:
                    timestamp, canid, frame = self._queue.get_nowait()
                assert isinstance(timestamp, Timestamp)
                assert isinstance(canid, CANID)
                assert isinstance(frame, UAVCANFrame)
            except (asyncio.TimeoutError, asyncio.QueueEmpty):
                # If there are unprocessed messages, allow the caller to read them even if the instance is closed.
                self._raise_if_closed()
                return None

            self._statistics.frames += 1

            if isinstance(canid, MessageCANID):
                assert isinstance(self._specifier.data_specifier, pyuavcan.transport.MessageDataSpecifier)
                assert self._specifier.data_specifier.subject_id == canid.subject_id
                source_node_id = canid.source_node_id
                if source_node_id is None:
                    # Anonymous transfer - no reconstruction needed
                    self._statistics.transfers += 1
                    self._statistics.payload_bytes += len(frame.padded_payload)
                    out = pyuavcan.transport.TransferFrom(
                        timestamp=timestamp,
                        priority=canid.priority,
                        transfer_id=frame.transfer_id,
                        fragmented_payload=[frame.padded_payload],
                        source_node_id=None,
                    )
                    _logger.debug("%s: Received anonymous transfer: %s; current stats: %s", self, out, self._statistics)
                    return out

            elif isinstance(canid, ServiceCANID):
                assert isinstance(self._specifier.data_specifier, pyuavcan.transport.ServiceDataSpecifier)
                assert self._specifier.data_specifier.service_id == canid.service_id
                assert (
                    self._specifier.data_specifier.role == pyuavcan.transport.ServiceDataSpecifier.Role.REQUEST
                ) == canid.request_not_response
                source_node_id = canid.source_node_id

            else:
                assert False

            receiver = self._receivers[source_node_id]
            result = receiver.process_frame(timestamp, canid.priority, frame, self._transfer_id_timeout_ns)
            if isinstance(result, TransferReassemblyErrorID):
                self._statistics.errors += 1
                self._statistics.reception_error_counters[result] += 1
                _logger.debug(
                    "%s: Rejecting CAN frame %s because %s; current stats: %s", self, frame, result, self._statistics
                )
            elif isinstance(result, pyuavcan.transport.TransferFrom):
                self._statistics.transfers += 1
                self._statistics.payload_bytes += sum(map(len, result.fragmented_payload))
                _logger.debug("%s: Received transfer: %s; current stats: %s", self, result, self._statistics)
                return result
            elif result is None:
                pass  # Nothing to do - expecting more frames
            else:
                assert False


def _node_id_range() -> typing.Iterable[int]:
    return range(CANID.NODE_ID_MASK + 1)


_NANO = 1e-9



# Copyright (C) 2021  UAVCAN Consortium  <uavcan.org>
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

from __future__ import annotations
import sys
import abc
from fnmatch import fnmatchcase
from typing import Optional, Iterator, Union, Callable, Tuple, Sequence, Dict
import logging
import pyuavcan
from . import backend
from ._value import RelaxedValue, ValueProxy, Value

if sys.version_info >= (3, 9):
    from collections.abc import MutableMapping
else:  # pragma: no cover
    from typing import MutableMapping  # pylint: disable=ungrouped-imports


class MissingRegisterError(KeyError):
    """
    Raised when the user attempts to access a register that is not defined.
    """


class ValueProxyWithFlags(ValueProxy):
    """
    This is like :class:`ValueProxy` but extended with register flags.
    """

    def __init__(self, msg: Value, mutable: bool, persistent: bool) -> None:
        super().__init__(msg)
        self._mutable = mutable
        self._persistent = persistent

    @property
    def mutable(self) -> bool:
        return self._mutable

    @property
    def persistent(self) -> bool:
        return self._persistent

    def __repr__(self) -> str:
        return pyuavcan.util.repr_attributes(self, repr(self.value), mutable=self.mutable, persistent=self.persistent)


class Registry(MutableMapping[str, ValueProxy]):
    """
    The registry (register repository) is the main access point for the application to its registers (configuration).
    It is a facade that provides user-friendly API on top of multiple underlying register backends
    (see :class:`backend.Backend`).
    Observe that it implements :class:`MutableMapping`.

    The user is not expected to instantiate this class manually;
    instead, it is provided as a member of :class:`pyuavcan.application.Node`,
    or via :func:`pyuavcan.application.make_node`.

    >>> import pyuavcan.application
    >>> registry = pyuavcan.application.make_registry(environment_variables={})

    Create static registers (stored in the register file):

    >>> from pyuavcan.application.register import Natural16, Real32
    >>> registry["p.a"] = Natural16([1234])                 # Assign or create.
    >>> registry.setdefault("p.b", Real32([12.34]))         # Update or create. # doctest: +NORMALIZE_WHITESPACE
    ValueProxyWithFlags(uavcan.register.Value...(real32=uavcan.primitive.array.Real32...(value=[12.34])),
                        mutable=True,
                        persistent=False)

    Create dynamic registers (getter/setter invoked at every access; existing entries overwritten automatically):

    >>> registry["d.a"] = lambda: [1.0, 2.0, 3.0]           # Immutable (read-only), deduced type: real64[3].
    >>> list(registry["d.a"].value.real64.value)            # Yup, deduced as expected, real64.
    [1.0, 2.0, 3.0]
    >>> registry["d.a"] = lambda: Real32([1.0, 2.0, 3.0])   # Like above, but now it is "real32[3]".
    >>> list(registry["d.a"].value.real32.value)
    [1.0, 2.0, 3.0]
    >>> d_b = [True, False, True]                   # Suppose we have some internal object.
    >>> def set_d_b(v: Value):                      # Define a setter for it.
    ...     global d_b
    ...     d_b = ValueProxy(v).bools
    >>> registry["d.b"] = (lambda: d_b), set_d_b    # Expose the object via mutable register with deduced type "bit[3]".

    Read/write/delete using the same dict-like API:

    >>> list(registry)  # Sorted lexicographically per backend. Altering backends affects register ordering.
    ['p.a', 'p.b', 'd.a', 'd.b']
    >>> len(registry)
    4
    >>> int(registry["p.a"])
    1234
    >>> registry["p.a"] = 88                        # Automatic type conversion to "natural16[1]" (defined above).
    >>> int(registry["p.a"])
    88
    >>> registry["d.b"].bools
    [True, False, True]
    >>> registry["d.b"] = [-1, 5, 0.0]              # Automatic type conversion to "bit[3]".
    >>> registry["d.b"].bools
    [True, True, False]
    >>> del registry["*.a"]                         # Use wildcards to remove multiple at the same time.
    >>> list(registry)
    ['p.b', 'd.b']
    >>> registry["d.b"].ints                        # Type conversion by ValueProxy.
    [1, 1, 0]
    >>> registry["d.b"].floats
    [1.0, 1.0, 0.0]
    >>> registry["d.b"].value.bit                   # doctest: +NORMALIZE_WHITESPACE
    uavcan.primitive.array.Bit...(value=[ True, True,False])

    Registers created by :meth:`setdefault` are always initialized from environment variables:

    >>> registry.environment_variables["P__C"] = b"999 +888.3"
    >>> registry.environment_variables["D__C"] = b"Hello world!"
    >>> registry.setdefault("p.c", Natural16([111, 222])).ints  # Value from environment is used here!
    [999, 888]
    >>> registry.setdefault("p.d", Natural16([111, 222])).ints  # No environment variable for this one.
    [111, 222]
    >>> d_c = 'Coffee'
    >>> def set_d_c(v: Value):
    ...     global d_c
    ...     d_c = str(ValueProxy(v))
    >>> str(registry.setdefault("d.c", (lambda: d_c, set_d_c))) # Setter is invoked immediately.
    'Hello world!'
    >>> registry["d.c"] = "New text"                            # Change the value again.
    >>> d_c                                                     # Yup, changed.
    'New text'
    >>> str(registry.setdefault("d.c", lambda: d_c))            # Environment var ignored because no setter.
    'New text'

    If such behavior is undesirable, one can either clear the environment variable dict or remove specific entries.
    See also: :func:`pyuavcan.application.make_node`.

    Variables created by direct assignment are (obviously) not affected by environment variables:

    >>> registry["p.c"] = [111, 222]                            # Direct assignment instead of setdefault().
    >>> registry["p.c"].ints                                    # Environment variables ignored!
    [111, 222]

    Closing the registry will close all underlying backends.

    >>> registry.close()

    TODO: Add modification notification callbacks to allow applications implement hot reloading.
    """

    Assignable = Union[
        RelaxedValue,
        Callable[[], RelaxedValue],
        Tuple[
            Callable[[], RelaxedValue],
            Callable[[Value], None],
        ],
    ]
    """
    An instance of any type from this union can be used to assign or create a register.
    Creation is handled depending on the type:

    - If a single callable, it will be invoked whenever this register is read; such register is called "dynamic".
      Such register will be reported as immutable.
      The registry file is not affected and therefore this change is not persistent.
      :attr:`environment_variables` are always ignored in this case since the register cannot be written.
      The result of the callable is converted to the register value using :class:`ValueProxy`.

    - If a tuple of two callables, then the first one is a getter that is invoked on read (see above),
      and the second is a setter that is invoked on write with a single argument of type :class:`Value`.
      It is guaranteed that the type of the value passed into the setter is always the same as that which
      is returned by the getter.
      The type conversion is performed automatically by polling the getter beforehand to discover the type.
      The registry file is not affected and therefore this change is not persistent.

    - Any other type (e.g., :class:`Value`, ``Natural16``, native, etc.):
      a static register will be created and stored in the registry file.
      Conversion logic is implemented by :class:`ValueProxy`.

    Dynamic registers (callables) overwrite existing entries unconditionally.
    It is not recommended to create dynamic registers with same names as existing static registers,
    as it may cause erratic behaviors.
    """

    @property
    @abc.abstractmethod
    def backends(self) -> Sequence[backend.Backend]:
        """
        If a register exists in more than one registry, only the first copy will be used;
        however, the count will include all redundant registers.
        """
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def environment_variables(self) -> Dict[str, bytes]:
        """
        When a new register is created using :meth:`setdefault`, its default value will be overridden from this dict.
        This is done to let the registry use values passed over to this node via environment variables or a similar
        mechanism.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def _create_static(self, name: str, value: Value) -> None:
        """This is an abstract method because only the implementation knows which backend should be used."""
        raise NotImplementedError

    @abc.abstractmethod
    def _create_dynamic(
        self,
        name: str,
        getter: Callable[[], Value],
        setter: Optional[Callable[[Value], None]],
    ) -> None:
        """This is an abstract method because only the implementation knows which backend should be used."""
        raise NotImplementedError

    def close(self) -> None:
        """
        Closes all storage backends.
        """
        for b in self.backends:
            b.close()

    def index(self, index: int) -> Optional[str]:
        """
        Get register name by index. The ordering is like :meth:`__iter__`. Returns None if index is out of range.
        """
        for i, key in enumerate(self):
            if i == index:
                return key
        return None

    def setdefault(self, name: str, default: Optional[Assignable] = None) -> ValueProxyWithFlags:
        """
        **This is the preferred method for creating new registers.**

        If the register exists, its value will be returned an no further action will be taken.
        If the register doesn't exist, it will be created and immediately updated from :attr:`environment_variables`
        (using :meth:`ValueProxy.assign_environment_variable`).
        The register value instance is created using :class:`ValueProxy`.

        :param name:    Register name.
        :param default: If exists, this value is ignored; otherwise created as described in :attr:`Assignable`.
        :return:        Resulting value.
        :raises:        See :meth:`ValueProxy.assign_environment_variable` and :meth:`ValueProxy`.
        """
        try:
            return self[name]
        except KeyError:
            pass
        if default is None:
            raise TypeError  # pragma: no cover
        from . import get_environment_variable_name

        _logger.debug("%r: Create %r <- %r", self, name, default)
        self._set(name, default, create_only=True)
        env_val = self.environment_variables.get(get_environment_variable_name(name))
        if env_val is not None:
            _logger.debug("%r: Update from env: %r <- %r", self, name, env_val)
            reg = self[name]
            reg.assign_environment_variable(env_val)
            self[name] = reg

        return self[name]

    def __getitem__(self, name: str) -> ValueProxyWithFlags:
        """
        :returns: :class:`ValueProxyWithFlags` (:class:`ValueProxy`) if exists.
        :raises: :class:`MissingRegisterError` (:class:`KeyError`) if no such register.
        """
        _ensure_name(name)
        for b in self.backends:
            ent = b.get(name)
            if ent is not None:
                return ValueProxyWithFlags(ent.value, mutable=ent.mutable, persistent=b.persistent)
        raise MissingRegisterError(name)

    def __setitem__(self, name: str, value: Assignable) -> None:
        """
        Assign a new value to the register if it exists and the type of the value is matching or can be
        converted to the register's type.
        The mutability flag may be ignored depending on which backend the register is stored at.
        The conversion is implemented by :meth:`ValueProxy.assign`.

        If the register does not exist, a new one will be created.
        However, unlike :meth:`setdefault`, :meth:`ValueProxy.assign_environment_variable` is not invoked.
        The register value instance is created using :class:`ValueProxy`.

        :raises:
            :class:`ValueConversionError` if the register exists but the value cannot be converted to its type
            or (in case of creation) the environment variable contains an invalid value.
        """
        self._set(name, value)

    def __delitem__(self, wildcard: str) -> None:
        """
        Remove registers that match the specified wildcard from all backends. Matching is case-sensitive.
        Count and keys are invalidated. **If no matching keys are found, no exception is raised.**
        """
        _ensure_name(wildcard)
        for b in self.backends:
            names = [n for n in b if fnmatchcase(n, wildcard)]
            _logger.debug("%r: Deleting %d registers matching %r from %r: %r", self, len(names), wildcard, b, names)
            for n in names:
                del b[n]

    def __iter__(self) -> Iterator[str]:
        """
        Iterator over register names. They may not be unique if different backends redefine the same register!
        The ordering is defined by backend ordering, then lexicographically.
        """
        return iter(n for b in self.backends for n in b.keys())

    def __len__(self) -> int:
        """
        Number of registers in all backends.
        """
        return sum(map(len, self.backends))

    def _set(self, name: str, value: Assignable, *, create_only: bool = False) -> None:
        _ensure_name(name)

        if callable(value):
            self._create_dynamic(name, lambda: ValueProxy(value()).value, None)  # type: ignore
            return
        if isinstance(value, tuple) and len(value) == 2 and all(map(callable, value)):
            g, s = value
            self._create_dynamic(name, (lambda: ValueProxy(g()).value), s)
            return

        if not create_only:
            for b in self.backends:
                e = b.get(name)
                if e is not None:
                    c = ValueProxy(e.value)
                    c.assign(value)
                    b[name] = c.value
                    return

        self._create_static(name, ValueProxy(value).value)

    def __repr__(self) -> str:
        return pyuavcan.util.repr_attributes(self, self.backends)


def _ensure_name(name: str) -> None:
    if not isinstance(name, str):
        raise TypeError(f"Register names are strings, not {type(name).__name__}")


_logger = logging.getLogger(__name__)



# Copyright (c) 2019 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

from ._base import SerialSession as SerialSession

from ._output import SerialOutputSession as SerialOutputSession
from ._output import SerialFeedback as SerialFeedback

from ._input import SerialInputSession as SerialInputSession
from ._input import SerialInputSessionStatistics as SerialInputSessionStatistics



# Copyright (c) 2019 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

import typing
import asyncio

import pyuavcan.transport


class LoopbackInputSession(pyuavcan.transport.InputSession):
    DEFAULT_TRANSFER_ID_TIMEOUT = 2

    def __init__(
        self,
        specifier: pyuavcan.transport.InputSessionSpecifier,
        payload_metadata: pyuavcan.transport.PayloadMetadata,
        loop: asyncio.AbstractEventLoop,
        closer: typing.Callable[[], None],
    ):
        self._specifier = specifier
        self._payload_metadata = payload_metadata
        self._loop = loop
        self._closer = closer
        self._transfer_id_timeout = float(self.DEFAULT_TRANSFER_ID_TIMEOUT)
        self._stats = pyuavcan.transport.SessionStatistics()
        self._queue: asyncio.Queue[pyuavcan.transport.TransferFrom] = asyncio.Queue()
        super().__init__()

    async def receive(self, monotonic_deadline: float) -> typing.Optional[pyuavcan.transport.TransferFrom]:
        timeout = monotonic_deadline - self._loop.time()
        try:
            if timeout > 0:
                out = await asyncio.wait_for(self._queue.get(), timeout)
            else:
                out = self._queue.get_nowait()
        except asyncio.TimeoutError:
            return None
        except asyncio.QueueEmpty:
            return None
        else:
            self._stats.transfers += 1
            self._stats.frames += 1
            self._stats.payload_bytes += sum(map(len, out.fragmented_payload))
            return out

    async def push(self, transfer: pyuavcan.transport.TransferFrom) -> None:
        """
        Inserts a transfer into the receive queue of this loopback session.
        """
        # TODO: handle Transfer ID like a real transport would: drop duplicates, handle transfer-ID timeout.
        # This is not very important for this demo transport but users may expect a more accurate modeling.
        await self._queue.put(transfer)

    @property
    def transfer_id_timeout(self) -> float:
        return self._transfer_id_timeout

    @transfer_id_timeout.setter
    def transfer_id_timeout(self, value: float) -> None:
        value = float(value)
        if value > 0:
            self._transfer_id_timeout = float(value)
        else:
            raise ValueError(f"Invalid TID timeout: {value!r}")

    @property
    def specifier(self) -> pyuavcan.transport.InputSessionSpecifier:
        return self._specifier

    @property
    def payload_metadata(self) -> pyuavcan.transport.PayloadMetadata:
        return self._payload_metadata

    def sample_statistics(self) -> pyuavcan.transport.SessionStatistics:
        return self._stats

    def close(self) -> None:
        self._closer()


def _unittest_session() -> None:
    import pytest

    closed = False

    specifier = pyuavcan.transport.InputSessionSpecifier(pyuavcan.transport.MessageDataSpecifier(123), 123)
    payload_metadata = pyuavcan.transport.PayloadMetadata(1234)

    def do_close() -> None:
        nonlocal closed
        closed = True

    ses = LoopbackInputSession(
        specifier=specifier, payload_metadata=payload_metadata, loop=asyncio.get_event_loop(), closer=do_close
    )

    ses.transfer_id_timeout = 123.456
    with pytest.raises(ValueError):
        ses.transfer_id_timeout = -0.1
    assert ses.transfer_id_timeout == pytest.approx(123.456)

    assert specifier == ses.specifier
    assert payload_metadata == ses.payload_metadata

    assert not closed
    ses.close()
    assert closed
